from .flattened_gaussian_profile import FlattenedGaussianTransverseProfile
from .gaussian_profile import GaussianTransverseProfile
from .hermite_gaussian_profile import HermiteGaussianTransverseProfile
from .jinc_profile import JincTransverseProfile
from .laguerre_gaussian_profile import LaguerreGaussianTransverseProfile
from .plane_wave_profile import PlaneWaveProfile
from .super_gaussian_profile import SuperGaussianTransverseProfile
from .transverse_profile import (
    ScaledTransverseProfile,
    SummedTransverseProfile,
    TransverseProfile,
)
from .transverse_profile_from_data import TransverseProfileFromData

__all__ = [
    "GaussianTransverseProfile",
    "HermiteGaussianTransverseProfile",
    "LaguerreGaussianTransverseProfile",
    "SuperGaussianTransverseProfile",
    "JincTransverseProfile",
    "TransverseProfileFromData",
    "TransverseProfile",
    "SummedTransverseProfile",
    "ScaledTransverseProfile",
    "PlaneWaveProfile",
    "FlattenedGaussianTransverseProfile",
]



# -*- coding: utf-8 -*-

import numpy as np
import pytest
from scipy.constants import c

from lasy.laser import Laser
from lasy.profiles.gaussian_profile import GaussianProfile
from lasy.utils.laser_utils import export_to_z, import_from_z


@pytest.fixture(scope="function")
def gaussian():
    # Cases with Gaussian laser
    wavelength = 0.8e-6
    pol = (1, 0)
    laser_energy = 1.0  # J
    t_peak = 0.0e-15  # s
    tau = 200.0e-15  # s
    w0 = 5.0e-6  # m
    profile = GaussianProfile(wavelength, pol, laser_energy, w0, tau, t_peak)

    return profile


def get_laser_z_analytic(profile, z_axis, r_axis):
    w0 = profile.w0
    tau = profile.tau
    omega0 = profile.omega0
    k0 = omega0 / c
    lambda0 = 2 * np.pi / k0

    L_Ray = np.pi * w0**2 / lambda0
    z_axis_2d = z_axis[None, :]
    r_axis_2d = r_axis[:, None]
    w0_z = w0 * np.sqrt(1 + (z_axis_2d / L_Ray) ** 2)
    R_z_inv = z_axis_2d / (z_axis_2d**2 + L_Ray**2)
    phi_gouy = np.arctan2(z_axis_2d, L_Ray)

    Field = (
        w0
        / w0_z
        * np.exp(-(r_axis_2d**2) / w0_z**2)
        * np.exp(-(z_axis_2d**2) / (c * tau) ** 2)
        * np.exp(1j * (k0 * r_axis_2d**2 * R_z_inv / 2 - phi_gouy))
    )

    return Field


def check_correctness(laser_t_in, laser_t_out, laser_z_analytic, z_axis):
    laser_z = export_to_z(laser_t_in.dim, laser_t_in.grid, laser_t_in.profile.omega0)
    import_from_z(
        laser_t_out.dim, laser_t_out.grid, laser_t_out.profile.omega0, laser_z, z_axis
    )

    field = laser_t_in.grid.get_temporal_field()
    ind0 = field.shape[0] // 2 - 1

    laser_t_in_2d = field[ind0]
    laser_t_out_2d = field[ind0]
    laser_z_2d = laser_z[ind0]

    assert np.allclose(laser_t_in_2d, laser_t_out_2d, atol=2e-7, rtol=0)

    assert np.allclose(laser_z_2d, laser_z_analytic, atol=1e-3, rtol=0)


def test_RT_case(gaussian):
    dim = "rt"
    w0 = gaussian.w0
    tau = gaussian.tau
    lo = (0, -3.5 * tau)
    hi = (5 * w0, 3.5 * tau)
    npoints = (128, 65)

    laser_t_in = Laser(dim, lo, hi, npoints, gaussian)
    laser_t_out = Laser(dim, lo, hi, npoints, gaussian)
    laser_t_in.normalize(1.0, "field")
    laser_t_out.normalize(1.0, "field")

    t_axis = laser_t_in.grid.axes[-1]
    r_axis = laser_t_in.grid.axes[0]
    z_axis = t_axis * c

    laser_z_analytic = get_laser_z_analytic(gaussian, z_axis, r_axis)

    check_correctness(laser_t_in, laser_t_out, laser_z_analytic, z_axis)


def test_3D_case(gaussian):
    # - 3D case
    dim = "xyt"
    w0 = gaussian.w0
    tau = gaussian.tau
    lo = (-5 * w0, -5 * w0, -3.5 * tau)
    hi = (5 * w0, 5 * w0, 3.5 * tau)
    npoints = (160, 164, 65)

    laser_t_in = Laser(dim, lo, hi, npoints, gaussian)
    laser_t_out = Laser(dim, lo, hi, npoints, gaussian)
    laser_t_in.normalize(1.0, "field")
    laser_t_out.normalize(1.0, "field")

    t_axis = laser_t_in.grid.axes[-1]
    r_axis = np.abs(laser_t_in.grid.axes[1])
    z_axis = t_axis * c

    laser_z_analytic = get_laser_z_analytic(gaussian, z_axis, r_axis)

    check_correctness(laser_t_in, laser_t_out, laser_z_analytic, z_axis)



import numpy as np
from scipy.constants import c

from .optical_element import OpticalElement


class Axiparabola(OpticalElement):
    r"""
    Class that represents an axiparabola.

    The rays that impinge on the axiparabola at different radii are focused
    to different positions on the axis (resulting in an extended "focal range").

    For more details, see S. Smartsev et al, "Axiparabola: a long-focal-depth,
    high-resolution mirror for broadband high-intensity lasers", Optics Letters 44, 14 (2019)

    Parameters
    ----------
    f0: float (in meter)
        The focal distance, i.e. the distance, from the axiparabola,
        where the focal range starts.

    delta: float (in meter)
        The length of the focal range.

    R: float (in meter)
        The radius of the axiparabola. Rays coming from ``r=0`` focus
        at ``z=f0`` ; rays coming from ``r=R`` focus at ``z=f0+delta``
    """

    def __init__(self, f0, delta, R):
        self.f0 = f0
        self.delta = delta
        self.R = R

    def amplitude_multiplier(self, x, y, omega):
        """
        Return the amplitude multiplier.

        Parameters
        ----------
        x, y, omega : ndarrays of floats
            Define points on which to evaluate the multiplier.
            These arrays need to all have the same shape.

        Returns
        -------
        multiplier : ndarray of complex numbers
            Contains the value of the multiplier at the specified points.
            This array has the same shape as the array omega.
        """
        # Implement Eq. 4 in Smatsev et al.
        r2 = x**2 + y**2
        sag = (
            (1.0 / (4 * self.f0)) * r2
            - (self.delta / (8 * self.f0**2 * self.R**2)) * r2**2
            + self.delta
            * (self.R**2 + 8 * self.f0 * self.delta)
            / (96 * self.f0**4 * self.R**4)
            * r2**3
        )

        # Calculate phase shift
        T = np.exp(-2j * (omega / c) * sag)
        # Remove intensity beyond R
        T[x**2 + y**2 > self.R**2] = 0

        return T



import numpy as np
from scipy.constants import c

from .longitudinal_profile import LongitudinalProfile


class LongitudinalProfileFromData(LongitudinalProfile):
    """
    Class for longitudinal laser profile created using data.

    The data used can either come from an experimental measurement
    or from the output of another code. This data is then used to
    define the longitudinal profile of the laser pulse.

    The data may be supplied in either the spectral or temporal
    domain. The data should be passed as a structure (defined
    below). If spectral data is passed, it will be converted to
    the temporal domain.

    Parameters
    ----------
    data : structure
        The data structure comprises several items indexed by
        a series of keys

        datatype : string
            The domain in which the data has been passed. Options
            are 'spectral' and 'temporal'.

        axis_is_wavelength : boolean (optional, default True)
            If True, the axis represents wavelength in [m] (SI).
            If False, it represents frequency in [1/s] (SI).

        axis : ndarrays of floats
            The horizontal axis of the pulse duration measurement.
            The array must be monotonically increasing or decreasing.
            When datatype is 'spectral' axis is wavelength in meters OR
            frequency in 1/seconds.
            When datatype is 'temporal' axis is time in seconds.

        intensity : ndarrays of floats
            The vertical axis of the pulse duration measurement.
            Spectral (resp. temporal) intensity when datatype is 'spectral' (resp 'temporal').

        phase : ndarray of floats
            If provided, this phase will be added to the pulse.
            When datatype is 'spectral' phase is spectral phase.
            When datatype is 'temporal' phase is temporal phase.

        dt : float
            Only required when datatype is 'spectral'. In this
            case this defines the user requested resolution in
            the conversion from the spectral to the temporal
            domain.

        wavelength : float
            Only required when datatype is 'temporal'. Then,
            this is the central wavelength of the pulse

    lo, hi : floats (seconds)
        Lower and higher ends of the required domain of the data.
        The data imported will be cut to this range prior to
        being incorporated into the ``lasy`` pulse.
    """

    def __init__(self, data, lo, hi):
        if data["datatype"] == "spectral":
            axis_is_wavelength = True  # Set to wavelength data by default
            if "axis_is_wavelength" in data:
                axis_is_wavelength = data["axis_is_wavelength"]
            if axis_is_wavelength:
                wavelength = data["axis"]  # Accept as wavelength
                spectral_intensity = data["intensity"]
            else:
                wavelength = 2.0 * np.pi * c / data["axis"]  # Convert to wavelength
                spectral_intensity = (
                    data["intensity"] * 2.0 * np.pi * c / wavelength**2
                )  # Convert spectral data
            assert np.all(np.diff(wavelength) > 0) or np.all(np.diff(wavelength) < 0), (
                'data["axis"] must be in monotonically increasing or decreasing order.'
            )
            if data.get("phase") is None:
                spectral_phase = np.zeros_like(wavelength)
            else:
                spectral_phase = data["phase"]
            if np.all(np.diff(wavelength) < 0):  # Flip arrays
                wavelength = wavelength[::-1]
                spectral_intensity = spectral_intensity[::-1]
                spectral_phase = spectral_phase[::-1]
            dt = data["dt"]
            cwl = np.sum(spectral_intensity * wavelength) / np.sum(spectral_intensity)
            cfreq = c / cwl
            # Determine required sampling frequency for desired dt
            sample_freq = 1 / dt
            # Determine number of points in temporal domain. This is the number of
            # points required to maintain the input spectral resolution while spanning
            # enough spectrum to achieve the desired temporal resolution.
            indx = np.argmin(np.abs(wavelength - cwl))
            dfreq = np.abs(c / wavelength[indx] - c / wavelength[indx + 1])
            N = int(sample_freq / dfreq)
            freq = np.linspace(cfreq - sample_freq / 2, cfreq + sample_freq / 2, N)
            # interpolate the spectrum onto this new array
            freq_intensity = np.interp(
                freq, c / wavelength[::-1], spectral_intensity[::-1], left=0, right=0
            )
            freq_phase = np.interp(
                freq, c / wavelength[::-1], spectral_phase[::-1], left=0, right=0
            )

            freq_amplitude = np.sqrt(freq_intensity)

            # Inverse Fourier Transform to the time domain
            t_amplitude = (
                np.fft.ifftshift(
                    np.fft.fft(
                        np.fft.fftshift(freq_amplitude * np.exp(1j * freq_phase))
                    )
                )
                / dt
            )
            time = np.linspace(-dt * N / 2, dt * N / 2 - dt, N)

            # Extract intensity and phase
            temporal_intensity = np.abs(t_amplitude) ** 2
            temporal_intensity /= np.max(temporal_intensity)
            temporal_phase = np.unwrap(-np.angle(t_amplitude))
            temporal_phase -= temporal_phase[np.argmin(np.abs(time))]

        elif data["datatype"] == "temporal":
            time = data["axis"]
            temporal_intensity = data["intensity"]
            if data.get("phase") is None:
                temporal_phase = np.zeros_like(time)
            else:
                temporal_phase = data["phase"]
            cwl = data["wavelength"]

        else:
            raise Exception("datatype must be 'spectral' or 'temporal'")

        super().__init__(cwl)

        # Finally crop the temporal domain to the physical domain
        # of interest

        tIndLo = np.argmin(np.abs(time - lo))
        tIndHi = np.argmin(np.abs(time - hi))

        self.time = time[tIndLo:tIndHi]
        self.temporal_intensity = temporal_intensity[tIndLo:tIndHi]
        self.temporal_phase = temporal_phase[tIndLo:tIndHi]

    def evaluate(self, t):
        """
        Return the longitudinal field envelope.

        Parameters
        ----------
        t : ndarray of floats
            Define points on which to evaluate the envelope

        Returns
        -------
        envelope : ndarray of complex numbers
            Contains the value of the longitudinal envelope at the
            specified points. This array has the same shape as the array t.
        """
        intensity = np.interp(t, self.time, self.temporal_intensity)
        phase = np.interp(t, self.time, self.temporal_phase)

        envelope = np.sqrt(intensity) * np.exp(-1j * phase)

        return envelope



import math

import numpy as np
from scipy.special import binom

from .transverse_profile import TransverseProfile


class FlattenedGaussianTransverseProfile(TransverseProfile):
    r"""
    Class for the analytic profile of a Flattened-Gaussian laser pulse.

    Define a complex transverse profile with a flattened Gaussian intensity
    distribution **far from focus** that transforms into a distribution
    with rings **in the focal plane**. (See `Santarsiero et al., J.
    Modern Optics, 1997 <http://doi.org/10.1080/09500349708232927>`_)

    Increasing the parameter ``N`` increases the
    flatness of the transverse profile **far from focus**,
    and increases the number of rings **in the focal plane**.

    The implementation of this class is based on that from `FBPIC`
    <https://github.com/fbpic/fbpic/blob/dev/fbpic/lpa_utils/laser/transverse_laser_profiles.py>.

    **In the focal plane** (:math:`z=0`), or in the far field, the profile translates to a
    laser with a transverse electric field:

    .. math::

        E(x,y,z=0) \propto
        \exp\left(-\frac{r^2}{(N+1)w_0^2}\right)
        \sum_{n=0}^N c'_n L^0_n\left(\frac{2\,r^2}{(N+1)w_0^2}\right)


    with Laguerre polynomials :math:`L^0_n` and

    .. math::

        \qquad c'_n=\sum_{m=n}^{N}\frac{1}{2^m}\binom{m}{n}

    - For :math:`N=0`, this is a Gaussian profile: :math:`E\propto\exp\left(-\frac{r^2}{w_0^2}\right)`.

    - For :math:`N\rightarrow\infty`, this is a Jinc profile: :math:`E\propto\frac{J_1(r/w_0)}{r/w_0}`.

    The equivalent expression for **the collimated beam** in the near field which produces this focus is
    given by:

    .. math::

        E(x,y) \propto
        \exp\left(-\frac{(N+1)r^2}{w^2}\right)
        \sum_{n=0}^N \frac{1}{n!}\left(\frac{(N+1)\,r^2}{w^2}\right)^n

    with the relationship between the spot sizes of the beams in the far field and in the near field given by:

    .. math::

        w = \frac{\lambda_0}{\pi w_0}|z_{\mathrm{foc}}|

    where :math:`z_{\mathrm{foc}}` is the distance between the far field and near field planes.

    - Note that a beam defined using the near field definition would be
      equivalent to a beam defined with the corresponding parameters far from focus in
      the far field, but without the parabolic phase arising from being
      defined far from the focus.

    - For :math:`N=0`, the near field profile is a Gaussian profile: :math:`E\propto\exp\left(-\frac{r^2}{w^2}\right)`.
    - For :math:`N\rightarrow\infty`, the near field profile is a flat profile: :math:`E\propto\Theta(w-r)`.

    Parameters
    ----------
    field_type : string
        Options: 'nearfield', when the beam is defined far from focus (e.g., right before the focusing optics), or 'farfield', when the beam is in the vicinity of the focus.

    w : float (in meter)
        The waist of the laser pulse. If ``field_type == 'farfield'`` then this
        variable corresponds to :math:`w_{0}` in the above far field formula.
        If ``field_type == 'nearfield'`` then this variable corresponds to
        :math:`w` in the above near field formula.

    N : int
        Determines the "flatness" of the transverse profile, far from
        focus (see the above formula).
        Default: ``N=6`` ; somewhat close to an 8th order supergaussian.

    wavelength : float (in meter)
        The main laser wavelength :math:`\lambda_0` of the laser.

    z_foc : float (in meter), optional
        Only required if defining the pulse in the far field. Gives the position
        of the focal plane. (The laser pulse is initialized at ``z=0``.)

    """

    def __init__(self, field_type, w, N, wavelength, z_foc=0):
        super().__init__()
        # Ensure that N is an integer
        assert isinstance(N, int) and N >= 0
        assert field_type in ["nearfield", "farfield"]
        self.field_type = field_type
        self.N = N

        if field_type == "farfield":
            w0 = w
            # Calculate effective waist of the Laguerre-Gauss modes, at focus
            self.w_foc = w0 * (self.N + 1) ** 0.5
            # Calculate Rayleigh Length
            self.zr = np.pi * self.w_foc**2 / wavelength
            # Evaluation distance w.r.t focal position
            self.z_eval = z_foc
            # Calculate the coefficients for the Laguerre-Gaussian modes
            self.cn = np.empty(self.N + 1)
            for n in range(self.N + 1):
                m_values = np.arange(n, self.N + 1)
                self.cn[n] = np.sum((1.0 / 2) ** m_values * binom(m_values, n)) / (
                    self.N + 1
                )
        else:
            self.w = w

    def _evaluate(self, x, y):
        """
        Return the transverse envelope.

        Parameters
        ----------
        x, y : ndarrays of floats
            Define points on which to evaluate the envelope
            These arrays need to all have the same shape.

        Returns
        -------
        envelope : ndarray of complex numbers
            Contains the value of the envelope at the specified points
            This array has the same shape as the arrays x, y
        """
        if self.field_type == "farfield":
            # Term for wavefront curvature + Gouy phase
            diffract_factor = 1.0 - 1j * self.z_eval / self.zr
            w = self.w_foc * np.abs(diffract_factor)
            psi = np.angle(diffract_factor)
            # Argument for the Laguerre polynomials
            scaled_radius_squared = 2 * (x**2 + y**2) / w**2

            # Sum recursively over the Laguerre polynomials
            laguerre_sum = np.zeros_like(x, dtype=np.complex128)
            for n in range(0, self.N + 1):
                # Recursive calculation of the Laguerre polynomial
                # - `L` represents $L_n$
                # - `L1` represents $L_{n-1}$
                # - `L2` represents $L_{n-2}$
                if n == 0:
                    L = 1.0
                elif n == 1:
                    L1 = L
                    L = 1.0 - scaled_radius_squared
                else:
                    L2 = L1
                    L1 = L
                    L = (((2 * n - 1) - scaled_radius_squared) * L1 - (n - 1) * L2) / n
                # Add to the sum, including the term for the additional Gouy phase
                laguerre_sum += self.cn[n] * np.exp(-(2j * n) * psi) * L

            # Final envelope: multiply by n-independent propagation factors
            exp_argument = -(x**2 + y**2) / (self.w_foc**2 * diffract_factor)
            envelope = laguerre_sum * np.exp(exp_argument) / diffract_factor

            return envelope

        else:
            N = self.N
            w = self.w

            sumseries = 0
            for n in range(N + 1):
                sumseries += (
                    1 / math.factorial(n) * ((N + 1) * (x**2 + y**2) / w**2) ** n
                )

            envelope = np.exp(-(N + 1) * (x**2 + y**2) / w**2) * sumseries

            return envelope



import numpy as np
from scipy.constants import c

from .optical_element import OpticalElement


class ParabolicMirror(OpticalElement):
    r"""
    Class for a parabolic mirror.

    More precisely, the amplitude multiplier corresponds to:

    .. math::

        T(\boldsymbol{x}_\perp,\omega) = \exp(-i\omega (x^2+y^2)/2cf)

    where
    :math:`\boldsymbol{x}_\perp` is the transverse coordinate (orthogonal
    to the propagation direction). The other parameters in this formula
    are defined below.

    Parameters
    ----------
    f : float (in meter)
        The focal length of the parabolic mirror.
    """

    def __init__(self, f):
        self.f = f

    def amplitude_multiplier(self, x, y, omega):
        """
        Return the amplitude multiplier.

        Parameters
        ----------
        x, y, omega : ndarrays of floats
            Define points on which to evaluate the multiplier.
            These arrays need to all have the same shape.

        Returns
        -------
        multiplier : ndarray of complex numbers
            Contains the value of the multiplier at the specified points.
            This array has the same shape as the array omega.
        """
        return np.exp(-1j * omega * (x**2 + y**2) / (2 * c * self.f))



import numpy as np


def find_center_of_mass(img):
    """
    Find the center of mass of an image.

    Parameters
    ----------
    img: 2Darray of floats
        The image on which to calculate the COM

    Returns
    -------
    x0 , y0: floats
        The center of mass of the image along the horizontal
        and the vertical. The units are in pixels.

    """
    rows, cols = np.shape(img)
    x = np.linspace(0, cols - 1, cols)
    y = np.linspace(0, rows - 1, rows)

    # find the beam center using COM
    img_tot = np.sum(img)
    x0 = np.sum(np.dot(img, x)) / img_tot
    y0 = np.sum(np.dot(img.T, y)) / img_tot

    return x0, y0


def find_d4sigma(img):
    """
    Find the D4Sigma measurement of the spot size in x and y in pixels.

    https://en.wikipedia.org/wiki/Beam_diameter#D4%CF%83_or_second-moment_width

    Parameters
    ----------
    img : A numpy array containing the spatial intensity profile of a laser pulse.

    Returns
    -------
    D4sigX : The D4sigma along the first (x) axis
    D4sigY : The D4sigma along the second (y) axis
    """
    rows, cols = np.shape(img)
    x = np.linspace(0, cols - 1, cols)
    y = np.linspace(0, rows - 1, rows)

    x0, y0 = find_center_of_mass(img)

    img_tot = np.sum(img)
    D4sigX = 4 * np.sqrt(np.sum(np.dot(img, (x - x0) ** 2)) / img_tot)
    D4sigY = 4 * np.sqrt(np.sum(np.dot(img.T, (y - y0) ** 2)) / img_tot)

    return D4sigX, D4sigY



# -*- coding: utf-8 -*-

import copy

import numpy as np
import pytest

from lasy.laser import Laser
from lasy.profiles.gaussian_profile import GaussianProfile
from lasy.utils.phase_retrieval import gerchberg_saxton_algo
from lasy.utils.zernike import zernike

w0 = 25.0e-6  # m


@pytest.fixture(scope="function")
def gaussian():
    # Cases with Gaussian laser
    wavelength = 0.8e-6
    pol = (1, 0)
    laser_energy = 1.0e-3  # J
    t_peak = 0.0  # s
    tau = 30.0e-15  # s
    profile = GaussianProfile(wavelength, pol, laser_energy, w0, tau, t_peak)

    return profile


def test_3D_case(gaussian):
    # - 3D case
    dim = "xyt"
    lo = (-75e-6, -75e-6, -50e-15)
    hi = (75e-6, 75e-6, 50e-15)
    npoints = (100, 100, 100)

    laser = Laser(dim, lo, hi, npoints, gaussian)

    # Add a phase aberration
    # CALCULATE THE REQUIRED PHASE ABERRATION
    x = np.linspace(lo[0], hi[0], npoints[0])
    y = np.linspace(lo[1], hi[1], npoints[1])
    X, Y = np.meshgrid(x, y)
    pupilRadius = 2 * w0
    phase = -0.2 * zernike(X, Y, (0, 0, pupilRadius), 3)

    R = np.sqrt(X**2 + Y**2)
    phaseMask = np.ones_like(phase)
    phaseMask[R > pupilRadius] = 0

    # NOW ADD THE PHASE TO EACH SLICE OF THE FOCUS
    phase3D = np.repeat(phase[:, :, np.newaxis], npoints[2], axis=2)
    field = laser.grid.get_temporal_field()
    laser.grid.set_temporal_field(np.abs(field) * np.exp(1j * phase3D))

    # PROPAGATE THE FIELD FIELD FOWARDS AND BACKWARDS BY 1 MM
    propDist = 2e-3
    laserForward = copy.deepcopy(laser)
    laserForward.propagate(propDist)
    laserBackward = copy.deepcopy(laser)
    laserBackward.propagate(-propDist)

    # PERFORM GERCHBERG-SAXTON ALGORTIHM TO RETRIEVE PHASE
    _, _, amp_error = gerchberg_saxton_algo(
        laserBackward,
        laserForward,
        2 * propDist,
        condition="max_iterations",
        max_iterations=100,
        debug=True,
    )

    assert amp_error < 1e-6



import h5py
import numpy as np
from scipy.constants import c

from .from_array_profile import FromArrayProfile


class FromInsightFile(FromArrayProfile):
    r"""
    Profile defined from insight measurement.

    Parameters
    ----------
    file_path : string
        Path to the file created by INSIGHT that contains the full field (e.g. Exyt_0.h5)

    pol : list of 2 complex numbers (dimensionless)
        Polarization vector. It corresponds to :math:`p_u` in the above
        formula ; :math:`p_x` is the first element of the list and
        :math:`p_y` is the second element of the list. Using complex
        numbers enables elliptical polarizations.

    omega0 : string or float
        Set the central frequency for the envelope construction. Can be a float value
        in [rad/s], or a string defining the method for automatic frequency detection:
        "barycenter" frequency is averaged over the power profile, "peak" frequency
        corresponding to the location of the maximum of the on-axis spectrum.
    """

    def __init__(self, file_path, pol, omega0="barycenter"):
        # read the data from H5 filed
        with h5py.File(file_path, "r") as hf:
            data = np.asanyarray(hf["data/Exyt_0"][()], dtype=np.complex128, order="C")
            t = np.asanyarray(hf["scales/t"][()], dtype=np.float64, order="C")
            x = np.asanyarray(hf["scales/x"][()], dtype=np.float64, order="C")
            y = np.asanyarray(hf["scales/y"][()], dtype=np.float64, order="C")

        # convert data and axes to SI units
        t *= 1e-15
        x *= 1e-3
        y *= 1e-3

        # get the field on axis and local frequencies
        field_onaxis = data[data.shape[0] // 2, data.shape[1] // 2, :]
        omega_array = -np.gradient(np.unwrap(np.angle(field_onaxis)), t)

        # choose the central frequency
        if omega0 == "peak":
            # using peak field frequency
            omega0 = omega_array[np.abs(field_onaxis).argmax()]
        elif omega0 == "barycenter":
            # or "center of mass" frequency
            omega0 = np.average(omega_array, weights=np.abs(field_onaxis) ** 2)
        else:
            assert isinstance(omega0, float)

        # check the complex field convention and correct if needed
        if omega0 < 0:
            omega0 *= -1
            data = np.conj(data)
            print("Warning: input field will be conjugated")

        # remove the envelope frequency
        data *= np.exp(1j * omega0 * t[None, None, :])

        # created LASY profile using FromArrayProfile class
        wavelength = 2 * np.pi * c / omega0
        dim = "xyt"
        axes = {"x": x, "y": y, "t": t}
        super().__init__(
            wavelength=wavelength,
            pol=pol,
            array=data,
            dim=dim,
            axes=axes,
            axes_order=["x", "y", "t"],
        )



from lasy.optical_elements.optical_element import OpticalElement


class IntensityMask(OpticalElement):
    """
    Class for a radially symmetric intensity mask.

    This creates an optical element which acts to mask out the intensity of a laser pulse.
    The mask is radially symmetric and can either mask intensity beyond a user defined radius, thus acting as
    an aperture. Alternatively, the optica can mask intensity within a user defined radius, acting in this case as
    an optic with a hole.

    Parameters
    ----------
    R : float (in meter)
        The radius of the mask
    center: tuple (floats)
        Center of the mask. Default is (0,0)
    mask_type: string
        Should be 'aperture' (default, allows light inside) or 'hole' (allows light outside).

    """

    def __init__(self, R, center=(0, 0), mask_type="aperture"):
        assert mask_type in ["aperture", "hole"], (
            "mask_type must be 'aperture' or 'hole'"
        )
        self.R = R
        self.center = center
        self.mask_type = mask_type

    def amplitude_multiplier(self, x, y, omega):
        """
        Return the amplitude multiplier.

        Parameters
        ----------
        x, y, omega : ndarrays of floats
            Define points on which to evaluate the multiplier.
            These arrays need to all have the same shape.

        Returns
        -------
        multiplier : ndarray of complex numbers
            Contains the value of the multiplier at the specified points.
            This array has the same shape as the array omega.
        """
        r_squared = (x - self.center[0]) ** 2 + (y - self.center[1]) ** 2
        mask = r_squared <= self.R**2  # True inside, False outside

        if self.mask_type == "aperture":
            return mask.astype(float)  # 1 inside, 0 outside
        else:  # "hole"
            return (~mask).astype(float)  # 0 inside, 1 outside



import numpy as np
from axiprop.containers import ScalarFieldEnvelope
from axiprop.lib import PropagatorFFT2, PropagatorResampling
from scipy.constants import c, e, epsilon_0, m_e
from scipy.interpolate import interp1d
from scipy.signal import hilbert

from .grid import Grid


def compute_laser_energy(dim, grid):
    """
    Compute the total laser energy that corresponds to the current envelope data.

    This is used mainly for normalization purposes.

    Parameters
    ----------
    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    grid : a Grid object.
        It contains an ndarray (V/m) with
        the value of the envelope field and the associated metadata
        that defines the points at which the laser is defined.

    Returns
    -------
    energy: float (in Joules)
    """
    # This uses the following volume integral:
    # $E_{laser} = \int dV \;\frac{\epsilon_0}{2} | E_{env} |^2$
    # which assumes that we can average over the oscilations at the
    # specified laser wavelength.
    # This probably needs to be generalized for few-cycle laser pulses.

    envelope = grid.get_temporal_field()

    dV = get_grid_cell_volume(grid, dim)

    if dim == "xyt":
        energy = ((dV * epsilon_0) * abs(envelope) ** 2).sum()
    else:  # dim == "rt":
        energy = (
            dV[np.newaxis, :, np.newaxis] * epsilon_0 * abs(envelope[:, :, :]) ** 2
        ).sum()

    if grid.is_envelope:
        energy *= 0.5

    return energy


def normalize_energy(dim, energy, grid):
    """
    Normalize energy of the laser pulse contained in grid.

    Normalizes by matching laser energy to an energy given as an input.

    Parameters
    ----------
    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    energy : scalar (J)
        Energy of the laser pulse after normalization.

    grid: a Grid object
        Contains value of the laser envelope and metadata.
    """
    if energy is None:
        return

    current_energy = compute_laser_energy(dim, grid)
    if current_energy == 0.0:
        print("Field is zero everywhere, normalization will be skipped")
    else:
        norm_factor = (energy / current_energy) ** 0.5
        field = grid.get_temporal_field()
        field *= norm_factor
        grid.set_temporal_field(field)


def normalize_peak_field_amplitude(amplitude, grid):
    """
    Normalize energy of the laser pulse contained in grid.

    Normalizes by matching laser field amplitude to a peak field amplitude given as an input.

    Parameters
    ----------
    amplitude : scalar (V/m)
        Peak field amplitude of the laser pulse after normalization.

    grid : a Grid object
        Contains value of the laser envelope and metadata.
    """
    if amplitude is not None:
        field = grid.get_temporal_field()
        field_max = np.abs(field).max()
        if field_max == 0.0:
            print("Field is zero everywhere, normalization will be skipped")
        else:
            field *= amplitude / field_max
            grid.set_temporal_field(field)


def normalize_peak_intensity(peak_intensity, grid):
    """
    Normalize energy of the laser pulse contained in grid.

    Normalizes by matching laser intensity to a peak intensity given as an input.

    Parameters
    ----------
    peak_intensity : scalar (W/m^2)
        Peak intensity of the laser pulse after normalization.

    grid : a Grid object
        Contains value of the laser envelope and metadata.
    """
    if peak_intensity is not None:
        field = grid.get_temporal_field()
        intensity = np.abs(epsilon_0 * field**2 / 2 * c)
        input_peak_intensity = intensity.max()
        if input_peak_intensity == 0.0:
            print("Field is zero everywhere, normalization will be skipped")
        else:
            field *= np.sqrt(peak_intensity / input_peak_intensity)
            grid.set_temporal_field(field)


def normalize_peak_fluence(peak_fluence, grid):
    """
    Normalize energy of the laser pulse contained in grid.

    Normalizes by matching laser fluence to a peak fluence given as an input.

    Parameters
    ----------
    peak_fluence : scalar (J/m^2)
        Peak fluence of the laser pulse after normalization.

    grid : a Grid object
        Contains value of the laser envelope and metadata.
    """
    if peak_fluence is not None:
        field = grid.get_temporal_field()
        fluence = get_laser_fluence(grid)
        input_peak_fluence = fluence.max()
        if input_peak_fluence == 0.0:
            print("Field is zero everywhere, normalization will be skipped")
        else:
            field *= np.sqrt(peak_fluence / input_peak_fluence)
            grid.set_temporal_field(field)


def normalize_average_intensity(average_intensity, grid):
    """
    Normalize energy of the laser pulse contained in grid.

    Normalizes by matching average laser intensity to an average intensity given as an input.

    Parameters
    ----------
    average_intensity : scalar (W/m^2)
        Average intensity of the laser pulse envelope after normalization.

    grid : a Grid object
        Contains value of the laser envelope and metadata.
    """
    if average_intensity is not None:
        field = grid.get_temporal_field()
        intensity = np.abs(epsilon_0 * field**2 / 2 * c)
        input_average_intensity = intensity.mean()
        if input_average_intensity == 0.0:
            print("Field is zero everywhere, normalization will be skipped")
        else:
            field *= np.sqrt(average_intensity / input_average_intensity)
            grid.set_temporal_field(field)


def normalize_peak_power(dim, peak_power, grid):
    """
    Normalize energy of the laser pulse contained in grid.

    Normalizes by matching laser power to a peak power given as an input.

    Parameters
    ----------
    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    peak_power : scalar (W)
        Peak power of the laser pulse after normalization.

    grid : a Grid object
        Contains value of the laser envelope and metadata.
    """
    if peak_power is not None:
        power = get_laser_power(dim, grid)
        input_peak_power = power.max()
        if input_peak_power == 0.0:
            print("Field is zero everywhere, normalization will be skipped")
        else:
            field = grid.get_temporal_field()
            grid.set_temporal_field(field * np.sqrt(peak_power / input_peak_power))


def get_laser_power(dim, grid):
    r"""
    Calculate the instantaneous power of the laser along the time axis.

    Parameters
    ----------
    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    grid : a Grid object.
        It contains an ndarray (V/m) with the value of the envelope field and the associated metadata that defines the points at which the laser is defined.

    Returns
    -------
    power : ndarray (W)
        The instantaneous laser power along the temporal axis.
    """
    field = grid.get_temporal_field()
    intensity = np.abs(epsilon_0 * field**2 / 2 * c)
    dz = grid.dx[-1] * c
    unit_area = get_grid_cell_volume(grid, dim) / dz
    power = intensity.sum(axis=tuple(range(intensity.ndim - 1))) * unit_area

    return power


def get_laser_fluence(grid):
    r"""
    Calculate the fluence of the laser in space.

    Parameters
    ----------
    grid : a Grid object.
        It contains an ndarray (V/m) with the value of the envelope field and the associated metadata that defines the points at which the laser is defined.

    Returns
    -------
    fluence : ndarray (J/m^2)
        The fluence of the laser pulse in space.
    """
    field = grid.get_temporal_field()
    intensity = np.abs(epsilon_0 * field**2 / 2 * c)
    fluence = np.squeeze(np.sum(intensity, axis=-1) * grid.dx[-1])

    return fluence


def get_full_field(laser, theta=0, slice=0, slice_axis="x", Nt=None):
    """
    Reconstruct the laser pulse with carrier frequency on the default grid.

    Parameters
    ----------
    theta : float (rad) (optional)
        Azimuthal angle
    slice : float (optional)
        Normalised position of the slice from -0.5 to 0.5.
    Nt: int (optional)
        Number of time points on which field should be sampled. If is None,
        the orignal time grid is used, otherwise field is interpolated on a
        new grid.

    Returns
    -------
        Et : ndarray (V/m)
            The reconstructed field, with shape (Nr, Nt) (for `rt`)
            or (Nx, Nt) (for `xyt`).
        extent : ndarray (Tmin, Tmax, Xmin, Xmax)
            Physical extent of the reconstructed field.
    """
    omega0 = laser.profile.omega0
    env = laser.grid.get_temporal_field()
    time_axis = laser.grid.axes[-1]

    # If the field is not an envelope, it is a full field, so no
    # reason to recompute the full field.
    assert laser.grid.is_envelope

    if laser.dim == "rt":
        azimuthal_phase = np.exp(-1j * laser.grid.azimuthal_modes * theta)
        env_upper = env * azimuthal_phase[:, None, None]
        env_upper = env_upper.sum(0)
        azimuthal_phase = np.exp(1j * laser.grid.azimuthal_modes * theta)
        env_lower = env * azimuthal_phase[:, None, None]
        env_lower = env_lower.sum(0)
        env = np.vstack((env_lower[::-1][:-1], env_upper))
    elif slice_axis == "x":
        Nx_middle = env.shape[0] // 2 - 1
        Nx_slice = int((1 + slice) * Nx_middle)
        env = env[Nx_slice, :]
    elif slice_axis == "y":
        Ny_middle = env.shape[1] // 2 - 1
        Ny_slice = int((1 + slice) * Ny_middle)
        env = env[:, Ny_slice, :]
    else:
        return None

    if Nt is not None:
        Nr = env.shape[0]
        time_axis_new = np.linspace(laser.grid.lo[-1], laser.grid.hi[-1], Nt)
        env_new = np.zeros((Nr, Nt), dtype=env.dtype)

        for ir in range(Nr):
            interp_fu_abs = interp1d(time_axis, np.abs(env[ir]))
            slice_abs = interp_fu_abs(time_axis_new)
            interp_fu_angl = interp1d(time_axis, np.unwrap(np.angle(env[ir])))
            slice_angl = interp_fu_angl(time_axis_new)
            env_new[ir] = slice_abs * np.exp(1j * slice_angl)

        time_axis = time_axis_new
        env = env_new

    env *= np.exp(-1j * omega0 * time_axis[None, :])
    env = np.real(env)

    if laser.dim == "rt":
        ext = np.array(
            [
                laser.grid.lo[-1],
                laser.grid.hi[-1],
                -laser.grid.hi[0],
                laser.grid.hi[0],
            ]
        )
    else:
        ext = np.array(
            [
                laser.grid.lo[-1],
                laser.grid.hi[-1],
                laser.grid.lo[0],
                laser.grid.hi[0],
            ]
        )

    return env, ext


def get_spectrum(
    grid, dim, range=None, bins=20, omega0=None, method="sum", ordering="zero_center"
):
    r"""
    Get the frequency spectrum of an envelope or electric field.

    The spectrum can be calculated in three different ways, depending on the
    `method` specified by the user:

    Initially, the spectrum is calculated as the Fourier transform of the
    electric field :math:`E(t)`.

    .. math::

        \int E(t) e^{-i \omega t} dt

    neglecting the negative frequencies. If ``method=="raw"``, no further
    processing is done and the returned spectrum is a complex array with the
    same transverse dimensions as the input grid. The units are
    :math:`\mathrm{V / Hz}`.

    For the other methods, the spectral energy density is calculated as

    .. math::

        \frac{\epsilon_0 c}{2\pi} |\int E(t) e^{-i \omega t} dt| ^ 2

    If ``method=="on_axis"``, a 1D real array with on-axis value of the
    equation above is returned. The units are :math:`\mathrm{J / (rad Hz m^2)}`.

    Otherwise, if ``method=="sum"`` (default), the transverse integral of the
    spectral energy density is calculated:

    .. math::

        \frac{\epsilon_0 c}{2\pi} \int |\int E(t) e^{-i \omega t} dt| ^ 2 dx dy

    The units of this array are :math:`\mathrm{J / (rad Hz)}`

    Parameters
    ----------
    grid : a Grid object.
        It contains an ndarray with the field data from which the
        spectrum is computed, and the associated metadata. The last axis must
        be the longitudinal dimension.

    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    range : list of float (optional)
        List of two values indicating the minimum and maximum frequency of the
        spectrum. If provided, only the FFT spectrum within this range
        will be returned using interpolation.

    bins : int (optional)
        Number of bins into which to interpolate the spectrum if a `range`
        is given.

    omega0 : scalar (optional)
        Angular frequency at which the envelope is defined.
        Only used if grid.is_envelope is True.

    method : {'sum', 'on_axis', 'raw'} (optional)
        Determines the type of spectrum that is returned as described above.
        By default 'sum'.

    ordering : string (optional)
        Order of the frequency array and corresponding spectrum.
        Options are:
        - ``"zero_center"``: np.fft.fftshift is applied so the frequency array is monotonous with 0 at the center.
        - ``"zero_first"``: The frequency array starts with positive frequencies, and negative frequencies are at the end. The array is not monotonous. This is the default with np.fft.ifft.

    Returns
    -------
    spectrum : ndarray
        Array with the spectrum (units and array type depend on ``method``).

    omega : ndarray
        Array with the angular frequencies of the spectrum.
    """
    spectral_field, omega = grid.get_spectral_field()
    # multiply by the number of points due to np.fft.fft normalization
    spectral_field *= grid.npoints[-1]

    # Get spectrum.
    if grid.is_envelope:
        # Assume that the FFT of the envelope and the FFT of the complex
        # conjugate of the envelope do not overlap. Then we only need
        # one of them.
        assert omega0 is not None
        spectrum = 0.5 * spectral_field * grid.dx[-1]
        if method == "on_axis":
            nx, ny, _ = spectrum.shape
            spectrum = spectrum[nx // 2, ny // 2] if dim == "xyt" else spectrum[0, 0]
        omega += omega0
    else:
        spectrum = spectral_field * grid.dx[-1]
        if method == "on_axis":
            nx, ny, _ = spectrum.shape
            spectrum = spectrum[nx // 2, ny // 2] if dim == "xyt" else spectrum[0, 0]

    # Convert to spectral energy density (J/(m^2 rad Hz)).
    if method != "raw":
        spectrum = np.abs(spectrum) ** 2 * epsilon_0 * c / np.pi

    # Integrate transversely.
    if method == "sum":
        dV = get_grid_cell_volume(grid, dim)
        dz = grid.dx[-1] * c
        if dim == "xyt":
            spectrum = np.sum(spectrum * dV / dz, axis=(0, 1))
        else:
            spectrum = np.sum(spectrum[0] * dV[:, np.newaxis] / dz, axis=0)

    assert ordering in ["zero_first", "zero_center"]
    if ordering == "zero_center":
        omega = np.fft.fftshift(omega, axes=-1)
        spectrum = np.fft.fftshift(spectrum, axes=-1)

    # If the user specified a frequency range, interpolate into it.
    if method in ["sum", "on_axis"] and range is not None:
        omega_interp = np.linspace(*range, bins)
        spectrum = np.interp(omega_interp, omega, spectrum)
        omega = omega_interp

    return spectrum, omega


def get_frequency(
    grid,
    dim=None,
    is_hilbert=False,
    omega0=None,
    phase_unwrap_nd=False,
    lower_bound=0.2,
    upper_bound=5.0,
):
    """
    Get the local and average frequency of a signal, either electric field or envelope.

    Parameters
    ----------
    grid : a Grid object.
        It contains a ndarrays with the field data from which the
        frequency is computed, and the associated metadata. The last axis must
        be the longitudinal dimension.
        Can be the full electric field or the envelope.

    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    is_hilbert : boolean (optional)
        If True, the field argument is assumed to be a Hilbert transform, and
        is used through the computation. Otherwise, the Hilbert transform is
        calculated in the function.

    omega0 : scalar
        Angular frequency at which the envelope is defined.
        Only used if grid.is_envelope is True.

    phase_unwrap_nd : boolean (optional)
        If True, the phase unwrapping is n-dimensional (2- or 3-D depending on dim).
        If False, the phase unwrapping is done in t, treating each transverse cell
        separately. This should be less accurate but faster.
        If set to True, scikit-image must be installed.

    lower_bound : scalar (optional)
        Relative lower bound for the local frequency
        Frequencies lower than lower_bound * central_omega are cut
        to lower_bound * central_omega.

    upper_bound : scalar (optional)
        Relative upper bound for the local frequency
        Frequencies larger than upper_bound * central_omega are cut
        to upper_bound * central_omega.

    Returns
    -------
    omega : nd array of doubles
        local angular frequency.

    central_omega : scalar
        Central angular frequency (averaged omega, weighted by the local
        envelope amplitude).
    """
    field = grid.get_temporal_field()

    # Assumes t is last dimension!
    if grid.is_envelope:
        assert omega0 is not None
        phase = np.unwrap(np.angle(field))
        omega = omega0 + np.gradient(-phase, grid.axes[-1], axis=-1, edge_order=2)
        central_omega = np.average(omega, weights=np.abs(field))
    else:
        assert dim in ["xyt", "rt"]
        if dim == "xyt" and phase_unwrap_nd:
            print("WARNING: using 3D phase unwrapping, this can be expensive")

        h = field if is_hilbert else hilbert_transform(field)

        if phase_unwrap_nd:
            try:
                from skimage.restoration import unwrap_phase

                skimage_installed = True
            except ImportError:
                skimage_installed = False
            assert skimage_installed, (
                "scikit-image must be install for nd phase unwrapping.",
                "Please install scikit-image or use phase_unwrap_nd=False.",
            )
            phase = unwrap_phase(np.angle(h))
        else:
            phase = np.unwrap(np.angle(h))
        omega = np.gradient(-phase, grid.axes[-1], axis=-1, edge_order=2)

        if dim == "xyt":
            weights = np.abs(h)
        else:
            r = grid.axes[0].reshape((grid.axes[0].size, 1))
            weights = r * np.abs(h)
        central_omega = np.average(omega, weights=weights)

    # Filter out too small frequencies
    omega = np.maximum(omega, lower_bound * central_omega)
    # Filter out too large frequencies
    omega = np.minimum(omega, upper_bound * central_omega)

    return omega, central_omega


def get_duration(grid, dim):
    """Get duration of the intensity of the envelope, measured as RMS.

    Parameters
    ----------
    grid : Grid
        The grid with the envelope to analyze.

    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    Returns
    -------
    float
        RMS duration of the envelope intensity in seconds.
    """
    # Calculate weights of each grid cell (amplitude of the field).
    dV = get_grid_cell_volume(grid, dim)
    field = grid.get_temporal_field()
    if dim == "xyt":
        weights = np.abs(field) ** 2 * dV
    else:  # dim == "rt":
        weights = np.abs(field) ** 2 * dV[np.newaxis, :, np.newaxis]
    # project weights to longitudinal axes
    weights = np.sum(weights, axis=(0, 1))
    return weighted_std(grid.axes[-1], weights)


def field_to_vector_potential(grid, omega0):
    """
    Convert envelope from electric field (V/m) to normalized vector potential.

    Parameters
    ----------
    grid : a Grid object.
        Contains the array of the electric field, to be converted to normalized
        vector potential, with corresponding metadata.
        The last axis must be the longitudinal dimension.

    omega0 : scalar
        Angular frequency at which the envelope is defined.

    Returns
    -------
    Normalized vector potential
    """
    # Here, we neglect the time derivative of the envelope of E, the first RHS
    # term in: E = -dA/dt + 1j * omega0 * A where E and A are the field and
    # vector potential envelopes, respectively
    assert grid.is_envelope
    omega, _ = get_frequency(grid, omega0=omega0)
    return -1j * e * grid.get_temporal_field() / (m_e * omega * c)


def vector_potential_to_field(grid, omega0, direct=True):
    """
    Convert envelope from electric field (V/m) to normalized vector potential.

    Parameters
    ----------
    grid : a Grid object.
        Contains the array of the normalized vector potential, to be
        converted to field, with corresponding metadata.
        The last axis must be the longitudinal dimension.

    omega0 : scalar
        Angular frequency at which the envelope is defined.

    direct : boolean (optional)
        If true, the conversion is done directly with derivative of vector
        potential. Otherwise, this is done using the local frequency.

    Returns
    -------
    Envelope of the electric field (V/m).
    """
    assert grid.is_envelope
    field = grid.get_temporal_field()
    if direct:
        A = (
            -np.gradient(field, grid.axes[-1], axis=-1, edge_order=2)
            + 1j * omega0 * field
        )
        return m_e * c / e * A
    else:
        omega, _ = get_frequency(grid, omega0=omega0)
        return 1j * m_e * omega * c * field / e


def field_to_envelope(grid, dim, phase_unwrap_nd=False):
    """Convert a field to its complex envelope representation by applying a Hilbert transform.

    Parameters
    ----------
    grid : Grid
        The Grid object on which the field is replaced with an envelope.
        This object is modified by the function.

    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    phase_unwrap_nd : boolean (optional)
        If True, the phase unwrapping is n-dimensional (2- or 3-D depending on dim).
        If False, the phase unwrapping is done in t, treating each transverse cell
        separately. This should be less accurate but faster.
        If set to True, scikit-image must be installed.

    Returns
    -------
    scalar
        The central angular frequency.
    """
    assert not grid.is_envelope

    # Get central wavelength from array
    _, omg0_h = get_frequency(
        grid,
        dim=dim,
        is_hilbert=False,
        phase_unwrap_nd=phase_unwrap_nd,
    )
    # Hilbert transform
    field = hilbert_transform(grid.get_temporal_field())
    # Remove carrier frequency
    field *= np.exp(1j * omg0_h * grid.axes[-1])
    # Store envelope
    grid.set_is_envelope(True)
    grid.set_temporal_field(field)

    return omg0_h


def hilbert_transform(field):
    """Make a hilbert transform of the field.

    Currently the arrays need to be flipped along t (both the input field and
    its transform) to get the imaginary part (and thus the phase) with the
    correct sign.

    Parameters
    ----------
    field : 3d numpy array
        The field whose field should be transformed.
    """
    return hilbert(field[:, :, ::-1])[:, :, ::-1]


def get_grid_cell_volume(grid, dim):
    """Get the volume of the grid cells.

    Parameters
    ----------
    grid : Grid
        The grid form which to compute the cell volume

    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    Returns
    -------
    float or ndarray
        A float with the cell volume (if dim=='xyt') or a numpy array with the
        radial distribution of cell volumes (if dim=='rt').
    """
    dz = grid.dx[-1] * c
    if dim == "xyt":
        dV = grid.dx[0] * grid.dx[1] * dz
    else:  # dim == "rt":
        r = grid.axes[0]
        dr = grid.dx[0]
        # 1D array that computes the volume of radial cells
        dV = np.pi * ((r + 0.5 * dr) ** 2 - (r - 0.5 * dr) ** 2) * dz
    return dV


def weighted_std(values, weights=None):
    """Calculate the weighted standard deviation of the given values.

    Parameters
    ----------
    values: array
        Contains the values to be analyzed

    weights : array
        Contains the weights of the values to analyze

    Returns
    -------
    A float with the value of the standard deviation
    """
    mean_val = np.average(values, weights=weights)
    std = np.sqrt(np.average((values - mean_val) ** 2, weights=weights))
    return std


def create_grid(array, axes, dim, is_envelope=True):
    """Create a lasy grid from a numpy array.

    Parameters
    ----------
    array : ndarray
        The input field array.

    axes : dict
        Dictionary with the information of the array axes.

    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    Returns
    -------
    grid : Grid
        A lasy grid containing the input array.
    """
    # Create grid.
    if dim == "xyt":
        lo = (axes["x"][0], axes["y"][0], axes["t"][0])
        hi = (axes["x"][-1], axes["y"][-1], axes["t"][-1])
        npoints = (axes["x"].size, axes["y"].size, axes["t"].size)
        grid = Grid(dim, lo, hi, npoints, is_envelope=is_envelope)
        assert np.allclose(grid.axes[0], axes["x"])
        assert np.allclose(grid.axes[1], axes["y"])
        assert np.allclose(grid.axes[2], axes["t"], rtol=1.0e-14)
        assert array.ndim == 3, "Input array should be of dimension 3 [x, y, time]"
        grid.set_temporal_field(array)
    else:  # dim == "rt":
        lo = (axes["r"][0], axes["t"][0])
        hi = (axes["r"][-1], axes["t"][-1])
        npoints = (axes["r"].size, axes["t"].size)
        nm = int((array.shape[0] + 1) / 2)
        grid = Grid(
            dim,
            lo,
            hi,
            npoints,
            n_azimuthal_modes=nm,
            is_envelope=is_envelope,
        )
        assert np.all(grid.axes[0] == axes["r"])
        assert np.allclose(grid.axes[1], axes["t"], rtol=1.0e-14)
        assert array.ndim == 3, (
            "Input array should be of dimension 3 [modes, radius, time]"
        )
        grid.set_temporal_field(array)
    return grid


def export_to_z(dim, grid, omega0, z_axis=None, z0=0.0, t0=0.0, backend="NP"):
    """
    Export laser pulse to spatial domain from temporal domain (internal LASY representation).

    Parameters
    ----------
    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    grid : a Grid object.
        It contains a ndarrays (V/m) with
        the value of the envelope field and the associated metadata
        that defines the points at which the laser is defined.

    omega0 : scalar
        Angular frequency at which the envelope is defined.

    z_axis : 1D ndarray of doubles (optional)
        Spatial `z` axis along which the field should be reconstructed.
        If not provided, `z_axis = c * t_axis` is considered.

    z0 : scalar (optional)
        Position from which the field is produced (emitted).

    t0 : scalar (optional)
        Moment of time at which the field is produced.

    backend : string (optional)
        Backend used by axiprop (see AVAILABLE_BACKENDS in axiprop
        documentation for more information).
    """
    time_axis_indx = -1

    t_axis = grid.axes[time_axis_indx]
    if z_axis is None:
        z_axis = t_axis * c

    FieldAxprp = ScalarFieldEnvelope(omega0 / c, t_axis)

    field = grid.get_temporal_field()

    if dim == "rt":
        # Construct the propagator
        prop = []
        for m in grid.azimuthal_modes:
            prop.append(
                PropagatorResampling(
                    grid.axes[0],
                    FieldAxprp.k_freq,
                    mode=m,
                    backend=backend,
                    verbose=False,
                )
            )

        field_z = np.zeros(
            (field.shape[0], field.shape[1], z_axis.size),
            dtype=field.dtype,
        )

        # Convert the spectral image to the spatial field representation
        for i_m in range(grid.azimuthal_modes.size):
            FieldAxprp.import_field(np.transpose(field[i_m]).copy())

            field_z[i_m] = prop[i_m].t2z(FieldAxprp.Field_ft, z_axis, z0=z0, t0=t0).T

            field_z[i_m] *= np.exp(-1j * (z_axis / c + t0) * omega0)
    else:
        # Construct the propagator
        Nx, Ny, Nt = field.shape
        Lx = grid.hi[0] - grid.lo[0]
        Ly = grid.hi[1] - grid.lo[1]
        prop = PropagatorFFT2(
            (Lx, Nx),
            (Ly, Ny),
            FieldAxprp.k_freq,
            backend=backend,
            verbose=False,
        )
        # Convert the spectral image to the spatial field representation
        FieldAxprp.import_field(np.moveaxis(field, -1, 0).copy())
        field_z = prop.t2z(FieldAxprp.Field_ft, z_axis, z0=z0, t0=t0)
        field_z = np.moveaxis(field_z, 0, -1)
        field_z *= np.exp(-1j * (z_axis / c + t0) * omega0)

    return field_z


def import_from_z(dim, grid, omega0, field_z, z_axis, z0=0.0, t0=0.0, backend="NP"):
    """
    Import laser pulse from spatial domain to temporal domain (internal LASY representation).

    Parameters
    ----------
    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    grid : a Grid object.
        It contains an ndarray (V/m) with
        the value of the envelope field and the associated metadata
        that defines the points at which the laser is defined.

    omega0 : scalar
        Angular frequency at which the envelope is defined.

    z_axis : 1D ndarray of doubles
        Spatial `z` axis along which the field should be reconstructed.

    z0 : scalar (optional)
        Position at which the field should be recorded

    t0 : scalar (optional)
        Moment of time at which the field should be recorded

    backend : string (optional)
        Backend used by axiprop (see AVAILABLE_BACKENDS in axiprop
        documentation for more information).
    """
    z_axis_indx = -1
    t_axis = grid.axes[z_axis_indx]
    dz = z_axis[1] - z_axis[0]
    Nz = z_axis.size

    # Transform the field from spatial to wavenumber domain
    field_fft = np.fft.fft(field_z, axis=z_axis_indx, norm="forward")

    # Create the axes for wavenumbers, and for corresponding frequency
    omega = 2 * np.pi * np.fft.fftfreq(Nz, dz / c) + omega0
    k_z = omega / c

    if dim == "rt":
        # Construct the propagator
        prop = []
        for m in grid.azimuthal_modes:
            prop.append(
                PropagatorResampling(
                    grid.axes[0],
                    omega / c,
                    mode=m,
                    backend=backend,
                    verbose=False,
                )
            )

        # Convert the spectral image to the spatial field representation
        field = np.zeros(grid.shape, dtype=np.complex128)
        for i_m in range(grid.azimuthal_modes.size):
            transform_data = np.transpose(field_fft[i_m]).copy()
            transform_data *= np.exp(-1j * z_axis[0] * (k_z[:, None] - omega0 / c))
            field[i_m] = prop[i_m].z2t(transform_data, t_axis, z0=z0, t0=t0).T
            field[i_m] *= np.exp(1j * (z0 / c + t_axis) * omega0)
        grid.set_temporal_field(field)
    else:
        # Construct the propagator
        Nx, Ny, _ = grid.npoints
        Lx = grid.hi[0] - grid.lo[0]
        Ly = grid.hi[1] - grid.lo[1]
        prop = PropagatorFFT2(
            (Lx, Nx),
            (Ly, Ny),
            omega / c,
            backend=backend,
            verbose=False,
        )
        # Convert the spectral image to the spatial field representation
        transform_data = np.moveaxis(field_fft, -1, 0).copy()
        transform_data *= np.exp(-1j * z_axis[0] * (k_z[:, None, None] - omega0 / c))
        field = np.moveaxis(prop.z2t(transform_data, t_axis, z0=z0, t0=t0), 0, -1)
        field *= np.exp(1j * (z0 / c + t_axis) * omega0)
        grid.set_temporal_field(field)


def get_w0(grid, dim):
    r"""
    Calculate the laser waist.

    Parameters
    ----------
    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    grid : a Grid object.
        It contains an ndarray (V/m) with the value of the envelope field and the associated metadata that defines the points at which the laser is defined.

    Returns
    -------
    sigma : Standard deviation of a**2 in m
    """
    field = grid.get_temporal_field()
    if dim == "xyt":
        Nx, Ny, Nt = field.shape
        A2 = (np.abs(field[Nx // 2 - 1, :, :]) ** 2).sum(-1)
        ax = grid.axes[1]
    else:
        A2 = (np.abs(field[0, :, :]) ** 2).sum(-1)
        ax = grid.axes[0]
        if ax[0] > 0:
            A2 = np.r_[A2[::-1], A2]
            ax = np.r_[-ax[::-1], ax]
        else:
            A2 = np.r_[A2[::-1][:-1], A2]
            ax = np.r_[-ax[::-1][:-1], ax]

    sigma = 2 * np.sqrt(np.average(ax**2, weights=A2))

    return sigma


def get_phi2(dim, grid):
    r"""
    Calculate the second derivative of the temporal phase of the laser.

    Parameters
    ----------
    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    grid : a Grid object.
        It contains an ndarray (V/m) with the value of the envelope field and the associated metadata that defines the points at which the laser is defined.

    Returns
    -------
    phi2 : Second derivative of temporal phase :math:`\Phi^{(2)} = \frac{d\omega_0}{dt} = \frac{d^2\Phi(t)}{dt^2}` in (second^-2)
    """
    env = grid.get_temporal_field()
    env_abs2 = np.abs(env**2)
    # Calculate group-delayed dispersion
    phi_envelop = np.unwrap(np.angle(env), axis=2)
    pphi_pt = np.gradient(phi_envelop, grid.dx[-1], axis=2)
    pphi_pt2 = np.gradient(pphi_pt, grid.dx[-1], axis=2)
    phi2 = np.average(pphi_pt2, weights=env_abs2)

    return phi2


def get_zeta(dim, grid, k0):
    r"""
    Calculate the spatial chirp of the laser.

    Parameters
    ----------
    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    grid : a Grid object.
        It contains an ndarray (V/m) with the value of the envelope field and the associated metadata that defines the points at which the laser is defined.

    Returns
    -------
    zeta_x, zeta_y : Spatial chirp in :math:`\zeta=\frac{dx_0}{d\omega}` (meter * second)
    nu_x, nu_y: Spatial chirp in :math:`\nu=\frac{d\omega_0}{dx}` (meter^-1 * second^-1)
    """
    assert dim == "xyt", "No spatial chirp for axis-symmetric dimension."
    w0 = get_w0(grid, dim)
    tau = 2 * get_duration(grid, dim)
    env_spec, spectral_axis = grid.get_spectral_field()
    env_spec_abs2 = np.abs(env_spec**2)
    # Get the spectral axis
    omega = spectral_axis + k0 * c
    # Calculate dx0 and dy0 in (x,y,omega) space
    weight_x_3d = np.transpose(env_spec_abs2, (2, 1, 0))
    weight_y_3d = np.transpose(env_spec_abs2, (2, 0, 1))
    weight_x_2d = np.sum(weight_x_3d, axis=2)
    weight_y_2d = np.sum(weight_y_3d, axis=2)
    # Calculate xda and yda, avoiding division by zero
    xda = np.where(
        weight_x_2d != 0, np.sum(grid.axes[0] * weight_x_3d, axis=2) / weight_x_2d, 0
    )
    yda = np.where(
        weight_y_2d != 0, np.sum(grid.axes[1] * weight_y_3d, axis=2) / weight_y_2d, 0
    )
    # Calculate spatial chirp zeta
    derivative_x_zeta = np.gradient(xda, omega, axis=0)
    derivative_y_zeta = np.gradient(yda, omega, axis=0)
    weight_x_2d = np.mean(env_spec_abs2, axis=0)
    weight_y_2d = np.mean(env_spec_abs2, axis=1)
    zeta_x = np.average(derivative_x_zeta.T, weights=weight_x_2d)
    zeta_y = np.average(derivative_y_zeta.T, weights=weight_y_2d)
    nu_x = 4 * zeta_x / (w0**2 * tau**2 + 4 * zeta_x**2)
    nu_y = 4 * zeta_y / (w0**2 * tau**2 + 4 * zeta_y**2)
    return [zeta_x, zeta_y], [nu_x, nu_y]


def get_beta(dim, grid, k0):
    r"""
    Calculate the angular dispersion of the laser.

    Parameters
    ----------
    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    grid : a Grid object.
        It contains an ndarray (V/m) with
        the value of the envelope field and the associated metadata that defines the points at which the laser is defined.

    Returns
    -------
    beta_x, beta_y : Angular dispersion in :math:` \beta = \frac{d\theta_0}{d\omega}` (second)
    """
    assert dim == "xyt", "No angular chirp for axis-symmetric dimension."
    env_spec, spectral_axis = grid.get_spectral_field()
    env_spec_abs2 = np.abs(env_spec**2)
    # Get the spectral axis
    omega = spectral_axis + k0 * c
    # Calculate angular dispersion beta
    phi_envelop_abs = np.unwrap(
        np.array(np.arctan2(env_spec.imag, env_spec.real)), axis=2
    )
    angle_x = np.gradient(phi_envelop_abs, grid.dx[1], axis=1) / k0
    angle_y = np.gradient(phi_envelop_abs, grid.dx[0], axis=0) / k0
    derivative_x_beta = np.gradient(angle_y, omega, axis=2)
    derivative_y_beta = np.gradient(angle_x, omega, axis=2)
    beta_x = np.average(derivative_x_beta, weights=env_spec_abs2)
    beta_y = np.average(derivative_y_beta, weights=env_spec_abs2)
    return [beta_x, beta_y]


def get_pft(dim, grid):
    r"""
    Calculate the pulse front tilt (PFT) of the laser.

    Parameters
    ----------
    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    grid : a Grid object.
        It contains an ndarray (V/m) with
        the value of the envelope field and the associated metadata
        that defines the points at which the laser is defined.

    Returns
    -------
    pft_x, pft_y : Pulse front tilt in :math:`p=\frac{dt_0}{dx}` (second * meter^-1).
    """
    assert dim == "xyt", "No pulse front tilt for cylindrical symmetry."
    env = grid.get_temporal_field()
    env_abs2 = np.abs(env**2)
    weight_xy_2d = np.mean(env_abs2, axis=2)
    z_centroids = np.sum(grid.axes[2] * env_abs2, axis=2) / np.sum(env_abs2, axis=2)
    derivative_x_pft = np.gradient(z_centroids, axis=0) / grid.dx[0]
    derivative_y_pft = np.gradient(z_centroids, axis=1) / grid.dx[1]
    pft_x = np.average(derivative_x_pft, weights=weight_xy_2d)
    pft_y = np.average(derivative_y_pft, weights=weight_xy_2d)
    return [pft_x, pft_y]


def get_propation_angle(dim, grid, k0):
    r"""
    Calculate the propagating angle of the laser.

    Parameters
    ----------
    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    grid : a Grid object.
        It contains an ndarray (V/m) with the value of the envelope field and the associated metadata that defines the points at which the laser is defined.

    Returns
    -------
    angle_x, angle_y : Propagating angle in :math:`p = \frac{k_x}{k_z}` or :math:`p = \frac{k_y}{k_z}` (in radians).
    """
    assert dim == "xyt", "Propagation is always on-axis for axis-symmetric dimension."
    env = grid.get_temporal_field()
    env_abs2 = np.abs(env**2)
    phi_envelop_abs = np.unwrap(np.angle(env), axis=2)
    pphi_px = np.gradient(phi_envelop_abs, grid.dx[1], axis=1)
    pphi_py = np.gradient(phi_envelop_abs, grid.dx[0], axis=0)
    angle_x = np.average(pphi_px, weights=env_abs2) / k0
    angle_y = np.average(pphi_py, weights=env_abs2) / k0
    return [angle_x, angle_y]


def get_spectral_phase(grid, dim, omega0, method="sum", ordering="zero_center"):
    r"""
    Calculate the spectral phase of a pulse in a given grid.

    Depending on the chosen calculation method, the spectral phase can be calculated in two different ways.

    If `method==sum` (default), the spectral field of the laser is spatially integrated before extracting the phase, i.e.

    .. math::

        \varphi(\omega) = \arg\left( \int E(x,y,\omega) dx dy \right)

    If `method==on-axis`, the on-axis spectral phase is calculated:

    .. math::

        \varphi(\omega) = \arg\left( E(x=0,y=0,\omega) \right)

    Parameters
    ----------
    grid : Grid
        The grid with the field to analyze.

    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    omega0 : float
        Central angular frequency of the field

    method : string, optional
        Method of calculating the spectral phase. Options are:

        - ``'sum'``: Calculates the spectral phase of the spatially summed field (default).
        - ``'on-axis'``: Calculates the on-axis spectral phase.

    ordering : string (optional)
        Order of the frequency array and corresponding spectral phase.
        Options are:
        - ``"zero_center"``: np.fft.fftshift is applied so the frequency array is monotonous with 0 at the center.
        - ``"zero_first"``: The frequency array starts with positive frequencies, and negative frequencies are at the end. The array is not monotonous. This is the default with np.fft.ifft.


    Returns
    -------
    phase: ndarray of floats (1D)
        Spectral phase of the pulse in the specified units and calculation method (in rad/s)

    omega: ndarray of floats (1D)
        Angular frequencies at which the phase is defined (in rad/s)

    """
    # Field must be envelope
    assert grid.is_envelope

    # get the spectral field
    field_spectral, omega = grid.get_spectral_field()

    # if method=='on-axis' get the on-axis field envelope, and calculate its phase
    assert method in ["on-axis", "sum"]
    if method == "on-axis":
        if dim == "xyt":
            Nx = grid.npoints[0]
            Ny = grid.npoints[1]
            phase = np.angle(field_spectral[Nx // 2, Ny // 2, :])
        else:  # dim=='rt'
            phase = np.angle(field_spectral[0, 0, :])

    # if method=='sum' integrate the field spatially before getting the phase from it
    else:  # method='sum'
        # grid cell volume required for integration
        dV = get_grid_cell_volume(grid, dim)

        if dim == "xyt":
            summed_field = np.sum(field_spectral * dV, axis=(0, 1))
        else:  # dim=='rt'
            summed_field = np.sum(field_spectral * dV[None, :, None], axis=(0, 1))

        phase = np.angle(summed_field)

    # create omega array (angular frequencies)
    assert ordering in ["zero_first", "zero_center"]
    if ordering == "zero_center":
        omega = np.fft.fftshift(omega, axes=-1)
        phase = np.fft.fftshift(phase, axes=-1)
    omega += omega0

    # unwrap the phase
    phase = np.unwrap(phase)

    # return the phase and omega arrays
    return phase, omega


def get_gdd(grid, dim, omega0, omega_gdd=None, method="sum"):
    r"""
    Calculate the group delay dispersion (GDD) of the laser.

    .. math::
        GDD = \frac{\partial^2 \phi(\omega)}{\partial \omega^2}


    Parameters
    ----------
    grid : Grid
        The grid with the field to analyze.

    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    omega0 : float
        Angular frequency at which the laser envelope is defined.

    omega_gdd : float, optional
        Central angular frequency at which the GDD is calculated, if `None`, `omega0` is used.

    method : string, optional
        Method of retrieving the phase that is used for calculating the GDD. Options are:

        - ``'sum'``: Calculates the spectral phase of the spatially summed field (default).
        - ``'on-axis'``: Calculates the on-axis spectral phase.

    Returns
    -------
    gdd: ndarray of floats (1D)
        Group delay dispersion over the entire spectral range (in s^2)

    gdd0: float
        Group delay dispersion at the center frequency (in s^2)

    """
    # calculate the spectral phase of the laser pulse
    phase, omega = get_spectral_phase(grid, dim, method=method, omega0=omega0)

    # calculate the second derivative wrt. angular frequency
    gd = np.gradient(phase, omega, axis=-1)
    gdd = np.gradient(gd, omega, axis=-1)

    # get the GDD at the specified frequency or the envelope's frequency
    omega_eval = omega_gdd if omega_gdd is not None else omega0
    gdd0 = np.interp(omega_eval, omega, gdd)
    return gdd, gdd0



import math

import numpy as np

from lasy.profiles.transverse.hermite_gaussian_profile import (
    HermiteGaussianTransverseProfile,
)
from lasy.profiles.transverse.transverse_profile import TransverseProfile
from lasy.utils.exp_data_utils import find_d4sigma


def hermite_gauss_decomposition(
    laserProfile,
    wavelength,
    res,
    lo,
    hi,
    m_max=12,
    n_max=12,
):
    """
    Decomposes a laser profile into a set of hermite-gaussian modes.

    The function takes an instance of `TransverseProfile`.

    Parameters
    ----------
    laserProfile : class instance
        An instance of a class or sub-class of TransverseLaserProfile

    wavelength : float (in meter)
        Central wavelength at which the Hermite-Gauss beams are to be defined.

    res : float
        The resolution of grid points in x and y that will be used
        during the decomposition calculation

    lo, hi : array of floats
        The lower and upper bounds of the spatial grid on which the
        decomposition will be performed.

    m_max, n_max : ints
        The maximum values of `m` and `n` up to which the expansion
        will be performed

    Returns
    -------
    weights : dict of floats
        A dictionary of floats corresponding to the weights of each mode
        in the decomposition. The keys of the dictionary are tuples
        corresponding to (`m`,`n`)

    w0x, w0y : floats
        Beam waist for which the decomposition is calculated.
        It is computed as the waist for which the weight of order 0 is maximum.
    """
    # Check if the provided laserProfile is a transverse profile.
    assert isinstance(laserProfile, TransverseProfile), (
        "laserProfile must be an instance of TransverseProfile"
    )

    # Get the field, sensible spatial bounds for the profile
    lo0 = lo[0] + laserProfile.x_offset
    lo1 = lo[1] + laserProfile.x_offset
    hi0 = hi[0] + laserProfile.y_offset
    hi1 = hi[1] + laserProfile.y_offset

    Nx = int((hi0 - lo0) // (2 * res) * 2) + 2
    Ny = int((hi1 - lo1) // (2 * res) * 2) + 2

    # Define spatial arrays
    x = np.linspace(
        (lo0 + hi0) / 2 - (Nx - 1) / 2 * res,
        (lo0 + hi0) / 2 + (Nx - 1) / 2 * res,
        Nx,
    )
    y = np.linspace(
        (lo1 + hi1) / 2 - (Ny - 1) / 2 * res,
        (lo1 + hi1) / 2 + (Ny - 1) / 2 * res,
        Ny,
    )
    X, Y = np.meshgrid(x, y)
    dx = x[1] - x[0]
    dy = y[1] - y[0]

    # Get the field on this grid
    field = laserProfile.evaluate(X, Y)

    # Get estimate of w0
    w0x, w0y = estimate_best_HG_waist(x, y, field, wavelength)

    # Next we loop over the modes and calculate the relevant weights
    weights = {}
    for m in range(m_max):
        for n in range(n_max):
            HGMode = HermiteGaussianTransverseProfile(w0x, w0y, m, n, wavelength)
            coef = np.real(
                np.sum(field * HGMode.evaluate(X, Y)) * dx * dy
            )  # modalDecomposition
            if math.isnan(coef):
                coef = 0
            weights[(m, n)] = coef

    return weights, w0x, w0y


def estimate_best_HG_waist(x, y, field, wavelength):
    """
    Estimate the waist that maximises the weighting of the first mode.

    Calculates a D4Sigma waist as a first estimate and then tests multiple
    gaussians with waists around this value to determine which has the best
    overlap with the provided intensity profile. The aim here is to maximise
    the energy in the fundamental mode of the reconstruction and so to avoid
    a decomposition with significant higher-order modal content.

    Parameters
    ----------
    x,y : 1D numpy arrays
        representing the x and y axes on which the intensity profile is defined.

    field : 2D numpy array representing the field (not the laser intensity).
        the laser field profile in a 2D slice.

    wavelength : float (in meter)
        Central wavelength at which the Hermite-Gauss beams are to be defined.

    Returns
    -------
    w0x, w0y : floats
        The calculated waist in x and y axis.
    """
    dx = x[1] - x[0]
    dy = y[1] - y[0]
    assert np.isclose(dx, dy, rtol=1e-10)

    X, Y = np.meshgrid(x, y)

    D4SigX, D4SigY = find_d4sigma(np.abs(field) ** 2)
    # convert this to a 1/e^2 width
    w0EstX = np.mean(D4SigX) / 2 * dx
    w0EstY = np.mean(D4SigY) / 2 * dy

    # Scan around the waist obtained from the D4sigma calculation,
    # and keep the waist for which this HG mode has the highest scalar
    # product with the input profile.
    waistTestX = np.linspace(w0EstX / 2, w0EstX * 1.5, 30)
    waistTestY = np.linspace(w0EstY / 2, w0EstY * 1.5, 30)
    coeffTest = np.zeros_like(waistTestX)

    for i in range(30):
        # create a gaussian
        HGMode = HermiteGaussianTransverseProfile(
            waistTestX[i], waistTestY[i], 0, 0, wavelength
        )
        profile = HGMode.evaluate(X, Y)
        coeffTest[i] = np.real(np.sum(profile * field))
    w0x = waistTestX[np.argmax(coeffTest)]
    w0y = waistTestY[np.argmax(coeffTest)]

    print("Estimated w0(x-axis) = %.2f microns (1/e^2 width)" % (w0x * 1e6))
    print("Estimated w0(y-axis) = %.2f microns (1/e^2 width)" % (w0y * 1e6))
    return w0x, w0y



from math import factorial

import numpy as np
from scipy.special import hermite

from .transverse_profile import TransverseProfile


class HermiteGaussianTransverseProfile(TransverseProfile):
    r"""
    A high-order Gaussian laser pulse expressed in the Hermite-Gaussian formalism.

    Definition is according to Siegman "Lasers" pg. 646 eq. 60, as explicitly given
    in https://doi.org/10.1364/JOSAB.489884 eq. 3, with the beam center upon the
    optical axis, :math:`x_0,y_0 = (0,0)`.

    More precisely, the transverse envelope (to be used in the
    :class:`.CombinedLongitudinalTransverseLaser` class) corresponds to:

    .. math::
        \mathcal{T}(x, y) = \,
                \mathcal{H}_m (x) \, \mathcal{H}_n(y) \, \exp(i \left ( \Phi_x(z) + \Phi_y(z)\right ) )

    with

    .. math::
        \mathcal{H}_p(q) = A_p h_p \left ( \frac{\sqrt{2}q}{w_q(z)} \right) \exp{\left( -\frac{q^2}{w_q^2(z)}\right)} \exp{ \left ( -i k_0 \frac{q^2}{2 R_q(z)} \right )}

        \Phi_q(z) = \left(p+\frac{1}{2}\right) \arctan\left({\frac{z}{Z_q}}\right)

        w_q(z) = w_{0,q} \sqrt{1 + \left( \frac{z}{Z_q}\right)^2}

        Z_q = \frac{\pi w_{0,q}^2}{\lambda_0}

        A_p = \frac{1}{\sqrt{w_q(z) 2^{p-1/2} p!\sqrt{\pi}}}

        R_q(z) = z + \frac{Z_q^2}{z}





    where  :math:`h_{p}` is the Hermite polynomial of order :math:`p`, :math:`w_q(z)` is the
    spot size of the laser along the :math:`q` axis (:math:`q` is :math:`x` or :math:`y`), :math:`Z_q` is the corresponding Rayleigh
    length and, :math:`\lambda_0` and :math:`k_0` are the wavelength and central wavenumber of the
    laser respectively.


    The z-dependence shown in the above equations is required to correctly define the
    electric field of the transverse profile relative to that of the pulse at the focus.
    The absolute z position will be overwritten when creating a laser object.

    Parameters
    ----------
    w_0x : float (in meter)
        The waist of the laser pulse in the x direction,
    w_0y : float (in meter)
        The waist of the laser pulse in the y direction,
    m : int (dimensionless)
        The order of hermite polynomial in the x direction
    n : int (dimensionless)
        The order of hermite polynomial in the y direction
    wavelength : float (in meter), optional
        The main laser wavelength :math:`\lambda_0` of the laser.
    z_foc : float (in meter), optional
        Position of the focal plane. (The laser pulse is initialized at
        ``z=0``.)

    Warnings
    --------
    In order to initialize the pulse out of focus, you can either:

    - Use a non-zero ``z_foc``
    - Use ``z_foc=0`` (i.e. initialize the pulse at focus) and then call
      ``laser.propagate(-z_foc)``

    Both methods are in principle equivalent, but note that the first
    method uses the paraxial approximation, while the second method does
    not make this approximation.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> from lasy.profiles.transverse.hermite_gaussian_profile import (
    ...     HermiteGaussianTransverseProfile,
    ... )
    >>> # Create evaluation grid
    >>> xy = np.linspace(-30e-6, 30e-6, 200)
    >>> X, Y = np.meshgrid(xy, xy)
    >>> # Create an array of plots
    >>> fig, ax = plt.subplots(3, 6, figsize=(10, 5), tight_layout=True)
    >>> extent = (1e6 * xy[0], 1e6 * xy[-1], 1e6 * xy[0], 1e6 * xy[-1])
    >>> for m in range(3):
    >>>     for n in range(3):
    >>>         transverse_profile = HermiteGaussianTransverseProfile(
    ...             w_0x = 10e-6, # m
    ...             w_0y = 15e-6, # m
    ...             m = m, #
    ...             n = n, #
    ...             wavelength = 0.8e-6, # m
    ...         )
    ...         intensity = np.abs(transverse_profile.evaluate(X,Y))**2
    ...         vmax_intensity = np.max(intensity)
    >>>         ax[m,n].imshow(intensity,extent=extent,cmap='bone_r',vmin=0,vmax=vmax_intensity)
    >>>         ax[m,n].set_title('Inten: m,n = %i,%i' %(m,n))
    >>>         phase = np.angle(transverse_profile.evaluate(X,Y))
    ...         vmax_phase = np.max(np.abs(phase))
    >>>         ax[m,n+3].imshow(phase,extent=extent,cmap='seismic',vmin=-vmax_phase,vmax=vmax_phase)
    >>>         ax[m,n+3].set_title('Phase: m,n = %i,%i' %(m,n))
    >>>         if m==2:
    >>>             ax[m,n].set_xlabel("x (µm)")
    >>>             ax[m,n+3].set_xlabel("x (µm)")
    >>>         else:
    >>>             ax[m,n].set_xticks([])
    >>>             ax[m,n+3].set_xticks([])
    >>>         if n==0:
    >>>             ax[m,n].set_ylabel("y (µm)")
    >>>             ax[m,n+3].set_yticks([])
    >>>         else:
    >>>             ax[m,n].set_yticks([])
    >>>             ax[m,n+3].set_yticks([])



    """

    def __init__(self, w_0x, w_0y, m, n, wavelength, z_foc=0):
        super().__init__()
        self.w_0x = w_0x
        self.w_0y = w_0y
        self.m = m
        self.n = n
        self.wavelength = wavelength
        self.z_foc = z_foc
        z_eval = -z_foc  # this links our observation position to Siegmann's definition

        self.k0 = 2 * np.pi / wavelength

        # Calculate Rayleigh Lengths
        Zx = np.pi * w_0x**2 / wavelength
        Zy = np.pi * w_0y**2 / wavelength

        # Calculate Size at Location Z
        wxZ = w_0x * np.sqrt(1 + (z_eval / Zx) ** 2)
        wyZ = w_0y * np.sqrt(1 + (z_eval / Zy) ** 2)

        # Calculate Multiplicative Factors
        Anx = 1 / np.sqrt(wxZ * 2 ** (m - 1 / 2) * factorial(m) * np.sqrt(np.pi))
        Any = 1 / np.sqrt(wyZ * 2 ** (n - 1 / 2) * factorial(n) * np.sqrt(np.pi))

        # Calculate the Phase contributions from propagation
        phiXz = (m + 1 / 2) * np.arctan2(z_eval, Zx)
        phiYz = (n + 1 / 2) * np.arctan2(z_eval, Zy)

        self.z_eval = z_eval
        self.Zx = Zx
        self.Zy = Zy
        self.wxZ = wxZ
        self.wyZ = wyZ
        self.Anx = Anx
        self.Any = Any
        self.phiXz = phiXz
        self.phiYz = phiYz

    def _evaluate(self, x, y):
        """
        Return the transverse envelope.

        Parameters
        ----------
        x, y : ndarrays of floats
            Define points on which to evaluate the envelope
            These arrays need to all have the same shape.

        Returns
        -------
        envelope : ndarray of complex numbers
            Contains the value of the envelope at the specified points
            This array has the same shape as the arrays x, y
        """
        z_eval = self.z_eval
        Zx = self.Zx
        Zy = self.Zy
        k0 = self.k0
        wxZ = self.wxZ
        wyZ = self.wyZ
        Anx = self.Anx
        Any = self.Any
        m = self.m
        n = self.n
        phiXz = self.phiXz
        phiYz = self.phiYz

        # Calculate the HG in each plane
        HGnx = (
            Anx
            * hermite(m)(np.sqrt(2) * (x) / wxZ)
            * np.exp(-((x) ** 2) / wxZ**2)
            * np.exp(-1j * k0 * (x) ** 2 / 2 / (z_eval**2 + Zx**2) * z_eval)
        )
        HGny = (
            Any
            * hermite(n)(np.sqrt(2) * (y) / wyZ)
            * np.exp(-((y) ** 2) / wyZ**2)
            * np.exp(-1j * k0 * (y) ** 2 / 2 / (z_eval**2 + Zy**2) * z_eval)
        )

        # Put it altogether
        envelope = HGnx * HGny * np.exp(1j * (phiXz + phiYz))

        return envelope



import numpy as np
from scipy.constants import c


class Profile(object):
    """
    Base class for all laser profiles.

    Any new laser profile should inherit from this class, and define its own
    `evaluate` method, using the same signature as the method below.
    For most cases, use derived classes instead of this base class.

    Common operators (addition and multiplication by a scalar) are provided as part of this base class.
    For such operations, the user is responsible for handling the complex phase and weights of summed profiles.
    In particular, summing between different types of profiles is not recommended.

    Parameters
    ----------
    wavelength : scalar
        Central wavelength for which the laser pulse envelope is defined.

    pol : list of 2 complex numbers
        Polarization vector that multiplies array_in to get the Ex and Ey fields.
        The envelope of each component of the electric field is given by:
        - Ex_env = array_in*pol(0)
        - Ey_env = array_in*pol(1)
        Standard polarizations can be obtained from:
        - Linear polarization in x: pol = (1,0)
        - Linear polarization in y: pol = (0,1)
        - Circular polarization: pol = (1,j)/sqrt(2) (j is the imaginary number)
        The polarization vector is normalized to have a unitary magnitude.

    """

    def __init__(self, wavelength, pol):
        assert len(pol) == 2
        norm_pol = np.sqrt(np.abs(pol[0]) ** 2 + np.abs(pol[1]) ** 2)
        self.pol = np.array([pol[0] / norm_pol, pol[1] / norm_pol])
        self.lambda0 = wavelength
        self.omega0 = 2 * np.pi * c / self.lambda0
        self.k0 = 2.0 * np.pi / wavelength
        self.is_cw = False
        self.is_plane_wave = False

    def evaluate(self, x, y, t):
        """
        Return the envelope field of the laser.

        Parameters
        ----------
        x, y, t : ndarrays of floats
            Define points on which to evaluate the envelope
            These arrays need to all have the same shape.

        Returns
        -------
        envelope : ndarray of complex numbers
            Contains the value of the envelope at the specified points
            This array has the same shape as the arrays x, y, t
        """
        # The base class only defines dummy fields
        # (This should be replaced by any class that inherits from this one.)
        return np.zeros_like(x, dtype="complex128")

    def __add__(self, other):
        """Return the sum of two profiles."""
        return SummedProfile(self, other)

    def __mul__(self, factor):
        """Return the scaled profile."""
        return ScaledProfile(self, factor)

    def __rmul__(self, factor):
        """Return the scaled profile."""
        return ScaledProfile(self, factor)

    def __update_is_cw__(self, value):
        """Update state of is_cw variable."""
        self.is_cw = value

    def __update_is_plane_wave__(self, value):
        """Update state of is_plane_wave variable."""
        self.is_plane_wave = value


class SummedProfile(Profile):
    """
    Base class for profiles that are the sum of several other profiles.

    Profile class that represents the sum of multiple profiles.

    Parameters
    ----------
    profiles : list of Profile objects
        List of profiles to be summed.
    """

    def __init__(self, *profiles):
        """Initialize the summed profile."""
        # Check that all profiles are Profile objects
        assert all([isinstance(p, Profile) for p in profiles]), (
            "All summands must be Profile objects."
        )
        self.profiles = profiles
        # Get the wavelength values from each profile
        lambda0s = [p.lambda0 for p in self.profiles]
        pols = [p.pol for p in self.profiles]
        # Check that all wavelengths are the same
        assert np.allclose(lambda0s, lambda0s[0]), (
            "Added profiles must have the same wavelength."
        )
        lambda0 = profiles[0].lambda0
        # Check that all polarizations are the same
        assert np.allclose(pols, pols[0]), (
            "Added profiles must have the same polarization."
        )
        pol = profiles[0].pol
        # Initialize the parent class
        super().__init__(lambda0, pol)

    def evaluate(self, x, y, t):
        """Return the envelope field of the summed profile."""
        # Sum the fields of each profile
        return sum([p.evaluate(x, y, t) for p in self.profiles])


class ScaledProfile(Profile):
    """
    Base class for profiles that are scaled by a factor.

    Profile class that represents scaled profiles.

    Parameters
    ----------
    profiles : Profile object
        Profile to be scaled.
    factor : int or float
        Factor by which to scale the profile.
    """

    def __init__(self, profile, factor):
        """Initialize the summed profile."""
        # Check that the factor is a number
        assert isinstance(factor, (int, float)), "The factor must be a number."
        # Check that the profile is a Profile object
        assert isinstance(profile, Profile), "The profile must be a Profile object."
        self.profile = profile
        self.factor = factor
        # Get the wavelength and polarization from the profile
        lambda0 = profile.lambda0
        pol = profile.pol
        # Initialize the parent class
        super().__init__(lambda0, pol)

    def evaluate(self, x, y, t):
        """Return the envelope field of the scaled profile."""
        # Sum the fields of each profile
        return self.profile.evaluate(x, y, t) * self.factor



import numpy as np

from .transverse_profile import TransverseProfile


class SuperGaussianTransverseProfile(TransverseProfile):
    r"""
    Class for the analytic profile of a super-Gaussian laser pulse.

    More precisely, the transverse envelope corresponds to:

    .. math::

        \mathcal{T}(x, y) = \exp\left( -\left({\frac{{x^2 + y^2}}{w_0^2}}\right)^{\dfrac{n}{2}} \right)

    Parameters
    ----------
    w0 : float (in meter)
        The waist of the laser pulse, i.e. :math:`w_0` in the above formula.

    n_order : float (in meter)
        The shape parameter of the super-gaussian function, i.e. :math:`n` in the above formula.
        If :math:`n=2` the super-Gaussian becomes a standard Gaussian function.
        If :math:`n=1` the super-Gaussian becomes a Laplace function.
    """

    def __init__(self, w0, n_order):
        super().__init__()
        self.w0 = w0
        self.n_order = n_order

    def _evaluate(self, x, y):
        """
        Return the transverse envelope.

        Parameters
        ----------
        x, y : ndarrays of floats
            Define points on which to evaluate the envelope
            These arrays need to all have the same shape.

        Returns
        -------
        envelope : ndarray of complex numbers
            Contains the value of the envelope at the specified points
            This array has the same shape as the arrays x, y
        """
        envelope = np.exp(-np.power((x**2 + y**2) / self.w0**2, self.n_order / 2))

        return envelope



import numpy as np
from scipy.constants import c, epsilon_0

from lasy.laser import Laser
from lasy.profiles.gaussian_profile import GaussianProfile
from lasy.utils.laser_utils import compute_laser_energy, get_duration, get_spectrum


def get_gaussian_profile():
    # Cases with Gaussian laser
    wavelength = 0.8e-6
    pol = (1, 0)
    laser_energy = 1.0  # J
    t_peak = 0.0e-15  # s
    tau = 30.0e-15  # s
    w0 = 5.0e-6  # m
    profile = GaussianProfile(wavelength, pol, laser_energy, w0, tau, t_peak)

    return profile


def get_gaussian_laser(dim):
    # - Cylindrical case
    if dim == "rt":
        lo = (0e-6, -60e-15)
        hi = (25e-6, +60e-15)
        npoints = (100, 100)
    else:  # dim == "xyt":
        lo = (-25e-6, -25e-6, -60e-15)
        hi = (+25e-6, +25e-6, +60e-15)
        npoints = (100, 100, 100)
    return Laser(dim, lo, hi, npoints, get_gaussian_profile())


def test_laser_analysis_utils():
    """Test the different laser analysis utilities in both geometries."""
    for dim in ["xyt", "rt"]:
        laser = get_gaussian_laser(dim)

        # Check that energy computed from spectrum agrees with `compute_laser_energy`.
        spectrum, omega = get_spectrum(laser.grid, dim, omega0=laser.profile.omega0)
        d_omega = omega[1] - omega[0]
        spectrum_energy = np.sum(spectrum) * d_omega
        energy = compute_laser_energy(dim, laser.grid)
        np.testing.assert_approx_equal(spectrum_energy, energy, significant=10)

        # Check that laser duration agrees with the given one.
        tau_rms = get_duration(laser.grid, dim)
        np.testing.assert_approx_equal(2 * tau_rms, laser.profile.tau, significant=3)


def test_laser_normalization_utils():
    """Test the different laser normalization utilities in both geometries."""
    for dim in ["xyt", "rt"]:
        laser = get_gaussian_laser(dim)

        # Check energy normalization
        laser.normalize(1, kind="energy")
        energy = compute_laser_energy(dim, laser.grid)
        np.testing.assert_approx_equal(1, energy, significant=10)

        # Check peak field normalization
        laser.normalize(1, kind="field")
        field = laser.grid.get_temporal_field()
        np.testing.assert_approx_equal(1, np.abs(field.max()), significant=10)

        # Check peak intensity normalization
        laser.normalize(1, kind="intensity")
        field = laser.grid.get_temporal_field()
        intensity = np.abs(epsilon_0 * field**2 / 2 * c)
        np.testing.assert_approx_equal(1, intensity.max(), significant=10)

        # Check average intensity normalization
        laser.normalize(1, kind="average_intensity")
        field = laser.grid.get_temporal_field()
        intensity = np.abs(epsilon_0 * field**2 / 2 * c)
        np.testing.assert_approx_equal(1, intensity.mean(), significant=10)


if __name__ == "__main__":
    test_laser_analysis_utils()
    test_laser_normalization_utils()



import math

import numpy as np


def get_zernike_nm(j):
    """
    Convert between different Zernike index schemes.

    Convert the OSA/ANSI Zernike Polynomial Index to the
    standard n,m indexing

    Parameters
    ----------
    j : int
        The OSA/ANSI Zernike Index

    Returns
    -------
    n,m : ints
        The standard Zernike Polynomial Indexes
    """
    n = int(np.ceil((-3 + np.sqrt(9 + 8 * j)) / 2))
    m = 2 * j - n * (n + 2)
    return int(m), int(n)


def zernike(x, y, pupil_coords, j):
    """
    Calculate the Zernike Polynomials to arbitrary order.

    Makes use of constructor formula on https://en.wikipedia.org/wiki/Zernike_polynomials

    Parameters
    ----------
    x, y : ndarrays (meters)
        The position at which to calculate the profile

    pupil_coords : tuple of floats (meters)
        A tuple of floats (cgx,cgy,r) with the first two elements corresponding to the center
        of the zernike mode and the third the radius of the mode

    j : int
        The OSA/ANSI Zernike Index

    Returns
    -------
    Z : ndarray (rad)
        The Zernike mode
    """
    # Setup
    (cgx, cgy, r) = pupil_coords
    rho = np.sqrt((x - cgx) ** 2 + (y - cgy) ** 2) / r
    theta = np.arctan2(y - cgy, x - cgx)

    m, n = get_zernike_nm(j)

    # next get the radial part
    R = RmnGenerator(n, abs(m), rho)

    # Now multiply by the azimuthal part
    if m < 0:
        Z = R * np.sin(-m * theta)
    else:
        Z = R * np.cos(m * theta)

    # Normalization
    if n == 0:
        scaling = 1
    else:
        if m == 0:
            scaling = np.sqrt((n + 1))
        else:
            scaling = np.sqrt(2 * (n + 1))
    Z = Z * scaling

    return Z


def RmnGenerator(n, m, rho):
    """
    Generate the Radial component of the Zernike Polynomials.

    Parameters
    ----------
    n,m : ints
        The standard Zernike Polynomial Indices

    rho : ndarray (meters)
        The radial positions at which to calculate the profile

    Returns
    -------
    Rmn : ndarray (rad)
        The radial component of the Zernike mode
    """
    if n == 0:
        if len(rho.shape) == 1:
            (r,) = rho.shape
            Rmn = np.ones(
                r,
            )
        else:
            r, c = rho.shape
            Rmn = np.ones((r, c))
    elif (n - m) % 2 == 0:
        # Even, Rmn is not 0
        k = np.linspace(0, int((n - m) / 2), int((n - m) / 2) + 1).astype(int)
        if len(rho.shape) == 1:
            (r,) = rho.shape
            Rmn = np.zeros(
                r,
            )
        else:
            r, c = rho.shape
            Rmn = np.zeros((r, c))
        for i in k:
            Rmn = Rmn + ((-1) ** i * math.factorial(n - i)) / (
                math.factorial(i)
                * math.factorial(int((n + m) / 2) - i)
                * math.factorial(int((n - m) / 2) - i)
            ) * rho ** (n - 2 * i)

    else:
        if len(rho.shape) == 1:
            (r,) = rho.shape
            Rmn = np.zeros(
                r,
            )
        else:
            r, c = rho.shape
            Rmn = np.zeros((r, c))

    return Rmn



import numpy as np

from .transverse_profile import TransverseProfile


class PlaneWaveProfile(TransverseProfile):
    r"""
    Class representing a plane wave.

    Specifically, the transverse profile will be represented by a constant value.

    Note that using this profile will force the number of points in each transverse
    dimension to 1. Additionally, it will force the area of the transverse plane to
    1 m^2.

    """

    def __init__(self):
        super().__init__()
        self.__update_is_plane_wave__(True)

    def _evaluate(self, x, y):
        """
        Return the transverse envelope.

        Parameters
        ----------
        x, y : ndarrays of floats
            Define points on which to evaluate the envelope
            These arrays need to all have the same shape.

        Returns
        -------
        envelope : ndarray of complex numbers
            Contains the value of the envelope at the specified points
            This array has the same shape as the arrays x, y
        """
        return np.ones_like(x**2 + y**2)



import numpy as np

from .optical_element import OpticalElement


class PolynomialSpectralPhase(OpticalElement):
    r"""
    Class for an optical element that adds spectral phase (e.g. a dazzler).

    The amplitude multiplier corresponds to:

    .. math::

        T(\omega) = \exp(i(\phi(\omega)))

    where :math:`\phi(\omega)` is the spectral phase given by:

    .. math::

        \phi(\omega) = \text{delay} (\omega - \omega_0) + \frac{\text{GDD}}{2!} (\omega - \omega_0)^2 + \frac{\text{TOD}}{3!} (\omega - \omega_0)^3 + \frac{\text{FOD}}{4!} (\omega - \omega_0)^4

    The other parameters in this formula are defined below.

    Parameters
    ----------
    omega0 : float (in rad/s)
        Central angular frequency about which the polynomial is expanded
    delay : float (in s), optional
        Group delay (by default: ``delay=0``). Positive value delays the pulse, i.e. it arrives at a later time
    gdd : float (in s^2), optional
        Group Delay Dispersion (by default: ``gdd=0``). ``gdd > 0`` corresponds to a positive
        chirp, i.e. the low-frequency part of the spectrum arriving earlier than the
        high-frequency part of the spectrum.
    tod : float (in s^3), optional
        Third-order Dispersion (by default: ``tod=0``). For a Gaussian pulse, adding a positive
        TOD (``tod > 0``) results in the apparition of post-pulses, i.e. lower intensity pulses
        arriving after the main pulse.
    fod : float (in s^4), optional
        Fourth-order Dispersion (by default: ``fod=0``).
    """

    def __init__(self, omega0, delay=0, gdd=0, tod=0, fod=0):
        self.omega0 = omega0
        self.delay = delay
        self.gdd = gdd
        self.tod = tod
        self.fod = fod

    def amplitude_multiplier(self, x, y, omega):
        """
        Return the amplitude multiplier.

        Parameters
        ----------
        x, y, omega : ndarrays of floats
            Define points on which to evaluate the multiplier.
            These arrays need to all have the same shape.


        Returns
        -------
        multiplier : ndarray of complex numbers
            Contains the value of the multiplier at the specified points.
            This array has the same shape as the array omega.
        """
        spectral_phase = (
            self.delay * (omega - self.omega0)
            + self.gdd / 2 * (omega - self.omega0) ** 2
            + self.tod / 6 * (omega - self.omega0) ** 3
            + self.fod / 24 * (omega - self.omega0) ** 4
        )

        return np.exp(1j * spectral_phase)



import numpy as np
from scipy.constants import c

from .optical_element import OpticalElement


class Axicon(OpticalElement):
    r"""
    Class for an axicon.

    This object technically represents a reflective axicon. However, it could
    also be used to represent a refractive axicon, if the chromatic effects of
    the refractive axicon are assumed to be negligible.

    More precisely, the amplitude multiplier corresponds to:

    .. math::

        T(\boldsymbol{x}_\perp,\omega) = \exp(-i (\omega/c) \sqrt{x^2+y^2} \tan(\gamma/2))

    where :math:`\boldsymbol{x}_\perp` is the transverse coordinate (orthogonal
    to the propagation direction). The other parameters in this formula
    are defined below.

    Parameters
    ----------
    gamma : float (in radians)
        The angle that the outcoming rays (coming from the axicon) would make with the optical axis,
        if the incoming rays (impinging on the axicon) are parallel to the optical axis.
    """

    def __init__(self, gamma):
        self.gamma = gamma

    def amplitude_multiplier(self, x, y, omega):
        """
        Return the amplitude multiplier.

        Parameters
        ----------
        x, y, omega : ndarrays of floats
            Define points on which to evaluate the multiplier.
            These arrays need to all have the same shape.

        Returns
        -------
        multiplier : ndarray of complex numbers
            Contains the value of the multiplier at the specified points.
            This array has the same shape as the array omega.
        """
        return np.exp(
            -2j * (omega / c) * np.sqrt(x**2 + y**2) * np.tan(0.5 * self.gamma)
        )



from .profile import Profile


class CombinedLongitudinalTransverseProfile(Profile):
    r"""
    Class that combines a longitudinal and transverse laser profile.

    The combined profile is defined as the product of the longitudinal and transverse
    profile.

    More precisely, the electric field corresponds to:

    .. math::

        E_u(\boldsymbol{x}_\perp,t) = Re\left[ E_0\, \mathcal{T}(x, y)
        \times \mathcal{L}(t) e^{-i\omega_0 t} \times p_u \right]

    where :math:`u` is either :math:`x` or :math:`y`, :math:`p_u` is
    the polarization vector, :math:`Re` represent the real part.
    The other parameters in this formula are defined below.

    Parameters
    ----------
    wavelength : float (in meter)
        The main laser wavelength :math:`\lambda_0` of the laser, which
        defines :math:`\omega_0` in the above formula, according to
        :math:`\omega_0 = 2\pi c/\lambda_0`.

    pol : list of 2 complex numbers (dimensionless)
        Polarization vector. It corresponds to :math:`p_u` in the above
        formula ; :math:`p_x` is the first element of the list and
        :math:`p_y` is the second element of the list. Using complex
        numbers enables elliptical polarizations.

    long_profile : :class:`.LongitudinalProfile`
        Defines the longitudinal envelope of the laser, i.e. the
        function :math:`\mathcal{L}(t)` in the above formula.

    transverse_profile : :class:`.TransverseProfile`
        Defines the transverse envelope of the laser, i.e. the
        function :math:`\mathcal{T}(x, y)` in the above formula.

    laser_energy : float (in Joule)
        The total energy of the laser pulse. The amplitude of the laser
        field (:math:`E_0` in the above formula) is automatically
        calculated so that the pulse has the prescribed energy.

    peak_fluence : float (in J/m^2)
        The peak fluence of the laser pulse. The amplitude of the laser
        field (:math:`E_0` in the above formula) is automatically
        calculated so that the pulse has the prescribed peak fluence.
        Required for a CW (monochromatic) profile.

    peak_power : float (in W)
        The peak power of the laser pulse. The amplitude of the laser
        field (:math:`E_0` in the above formula) is automatically
        calculated so that the pulse has the prescribed peak power.
        Required for a plane wave profile.
    """

    def __init__(
        self,
        wavelength,
        pol,
        long_profile,
        trans_profile,
        laser_energy=None,
        peak_fluence=None,
        peak_power=None,
    ):
        super().__init__(wavelength, pol)
        assert (
            (laser_energy is not None)
            ^ (peak_fluence is not None)
            ^ (peak_power is not None)
        ), "Exactly one of laser_energy, peak_fluence, or peak_power must be specified"
        if long_profile.is_cw:
            assert peak_fluence is not None
            self.peak_fluence = peak_fluence
            self.__update_is_cw__(long_profile.is_cw)

        elif trans_profile.is_plane_wave:
            assert peak_power is not None
            self.peak_power = peak_power
            self.__update_is_plane_wave__(trans_profile.is_plane_wave)
        else:
            assert laser_energy is not None
            self.laser_energy = laser_energy
        self.long_profile = long_profile
        self.trans_profile = trans_profile

    def evaluate(self, x, y, t):
        """
        Return the envelope field of the laser.

        Parameters
        ----------
        x, y, t : ndarrays of floats
            Define points on which to evaluate the envelope
            These arrays need to all have the same shape.

        Returns
        -------
        envelope : ndarray of complex numbers
            Contains the value of the envelope at the specified points
            This array has the same shape as the arrays x, y, t
        """
        envelope = self.trans_profile.evaluate(x, y) * self.long_profile.evaluate(t)
        return envelope



import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.constants import c, epsilon_0

from .laser_utils import get_duration, get_w0


def show_laser(grid, dim, show_intensity, **kw):
    """
    Show a 2D image of the laser represented on the grid.

    Parameters
    ----------
    grid : Grid
        The Grid object to be plotted

    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    show_intensity : bool
        if False the laser amplitude is plotted
        if True then the intensity of the laser is plotted along with lineouts
        and a measure of the pulse duration and spot size

    **kw : additional arguments to be passed to matplotlib's imshow command
    """
    if show_intensity:
        F = epsilon_0 * c / 2 * np.abs(grid.get_temporal_field()) ** 2 / 1e4
        cbar_label = r"I (W/cm$^2$)"
    else:
        F = np.abs(grid.get_temporal_field())
        cbar_label = r"$|E_{envelope}|$ (V/m)"

    # Calculate spatial scales for the axes
    if grid.hi[0] > 1:
        # scale is meters
        spatial_scale = 1
        spatial_unit = r"(m)"
    elif grid.hi[0] > 1e-3:
        # scale is millimeters
        spatial_scale = 1e-3
        spatial_unit = r"(mm)"
    else:
        # scale is microns
        spatial_scale = 1e-6
        spatial_unit = r"($\mu m$)"

    # Calculate temporal scales for the axes
    if grid.hi[-1] - grid.lo[-1] > 1e-9:
        # scale is nanoseconds
        temporal_scale = 1e-9
        temporal_unit = r"(ns)"
    elif grid.hi[-1] - grid.lo[-1] > 1e-12:
        # scale is picoseconds
        temporal_scale = 1e-12
        temporal_unit = r"(ps)"
    else:
        # scale is femtoseconds
        temporal_scale = 1e-15
        temporal_unit = r"(fs)"

    if dim == "rt":
        # Show field in the plane y=0, above and below axis, with proper sign for each mode
        F_plot = [
            np.concatenate(((-1.0) ** m * F[m, ::-1], F[m]))
            for m in grid.azimuthal_modes
        ]
        F_plot = sum(F_plot)  # Sum all the modes
        extent = [
            grid.lo[-1] / temporal_scale,
            grid.hi[-1] / temporal_scale,
            -grid.hi[0] / spatial_scale,
            grid.hi[0] / spatial_scale,
        ]

    else:
        # In 3D show an image in the xt plane
        i_slice = int(F.shape[1] // 2)
        F_plot = F[:, i_slice, :]
        extent = [
            grid.lo[-1] / temporal_scale,
            grid.hi[-1] / temporal_scale,
            grid.lo[0] / spatial_scale,
            grid.hi[0] / spatial_scale,
        ]

    fig, ax = plt.subplots()
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    im = ax.imshow(
        F_plot, extent=extent, cmap="Reds", aspect="auto", origin="lower", **kw
    )
    cb = fig.colorbar(im, cax=cax)
    cb.set_label(cbar_label)
    ax.set_xlabel(r"t " + temporal_unit)
    ax.set_ylabel(r"x " + spatial_unit)

    if show_intensity:
        # Create projected lineouts along time and space
        temporal_lineout = np.sum(F_plot, axis=0) / np.sum(F_plot, axis=0).max()
        ax.plot(
            grid.axes[-1] / temporal_scale,
            0.15 * temporal_lineout * (extent[3] - extent[2]) + extent[2],
            c=(0.3, 0.3, 0.3),
        )

        spatial_lineout = np.sum(F_plot, axis=1) / np.sum(F_plot, axis=1).max()
        ax.plot(
            0.15 * spatial_lineout * (extent[1] - extent[0]) + extent[0],
            np.linspace(extent[2], extent[3], F_plot.shape[0]),
            c=(0.3, 0.3, 0.3),
        )

        # Get the pulse duration
        tau = 2 * get_duration(grid, dim) / temporal_scale
        ax.text(
            0.55,
            0.95,
            r"Pulse Duration   = %.1f " % (tau) + temporal_unit[1:-1],
            transform=ax.transAxes,
        )

        # Get the spot size
        w0 = get_w0(grid, dim) / spatial_scale
        ax.text(
            0.55,
            0.9,
            r"Spot Size           = %.1f " % (w0) + spatial_unit[1:-1],
            transform=ax.transAxes,
        )



from .combined_profile import CombinedLongitudinalTransverseProfile
from .from_array_profile import FromArrayProfile
from .from_insight_file import FromInsightFile
from .from_openpmd_profile import FromOpenPMDProfile
from .gaussian_profile import GaussianProfile
from .profile import Profile
from .speckle_profile import SpeckleProfile

__all__ = [
    "Profile",
    "CombinedLongitudinalTransverseProfile",
    "GaussianProfile",
    "FromArrayProfile",
    "FromOpenPMDProfile",
    "FromInsightFile",
    "SpeckleProfile",
]



import os

import numpy as np
import openpmd_api as io
from scipy.constants import c

from lasy import __version__ as lasy_version

from .laser_utils import field_to_vector_potential


def write_to_openpmd_file(
    dim,
    write_dir,
    file_prefix,
    file_format,
    iteration,
    grid,
    wavelength,
    pol,
    save_as_vector_potential=False,
):
    """
    Write the laser field into an openPMD file.

    Parameters
    ----------
    dim : string
        Dimensionality of the array. Options are:

        - 'xyt': The laser pulse is represented on a 3D grid:
                 Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - 'rt' : The laser pulse is represented on a 2D grid:
                 Cylindrical (r) transversely, and temporal (t) longitudinally.

    file_prefix : string
        The file name will start with this prefix.

    write_dir : string
        The directory where the file will be written.

    file_format : string
        Format to be used for the output file. Options are "h5" and "bp".

    iteration : int
        The iteration number for the file to be written.

    grid : Grid
        A grid object containing the 3-dimensional array
        with complex envelope of the electric field and metadata

    wavelength : scalar
        Central wavelength for which the laser pulse envelope is defined.

    pol : list of 2 complex numbers
        Polarization vector that multiplies array to get the Ex and Ey arrays.

    save_as_vector_potential : bool (optional)
        Whether the envelope is converted to normalized vector potential
        before writing to file.
    """
    array = grid.get_temporal_field()

    # Create file
    full_filepath = os.path.join(
        write_dir, "{}_%05T.{}".format(file_prefix, file_format)
    )
    os.makedirs(write_dir, exist_ok=True)
    series = io.Series(full_filepath, io.Access.create)
    series.set_software("lasy", lasy_version)

    i = series.iterations[iteration]

    # Define the mesh
    m = i.meshes["laserEnvelope"]
    m.grid_spacing = [
        (hi - lo) / (npoints - 1)
        for hi, lo, npoints in zip(grid.hi, grid.lo, grid.npoints)
    ][::-1]
    m.grid_global_offset = grid.lo[::-1]
    if dim == "xyt":
        m.geometry = io.Geometry.cartesian
        m.axis_labels = ["t", "y", "x"]
    elif dim == "rt":
        m.geometry = io.Geometry.thetaMode
        m.axis_labels = ["t", "r"]

    # Store metadata needed to reconstruct the field
    m.set_attribute("angularFrequency", 2 * np.pi * c / wavelength)
    m.set_attribute("polarization", pol)
    if save_as_vector_potential:
        m.set_attribute("envelopeField", "normalized_vector_potential")
        m.unit_dimension = {}
    else:
        m.set_attribute("envelopeField", "electric_field")
        m.unit_dimension = {
            io.Unit_Dimension.M: 1,
            io.Unit_Dimension.L: 1,
            io.Unit_Dimension.I: -1,
            io.Unit_Dimension.T: -3,
        }

    if save_as_vector_potential:
        array = field_to_vector_potential(grid, 2 * np.pi * c / wavelength)

    # Pick the correct field
    assert dim in ["xyt", "rt"]
    if dim == "xyt":
        # Switch from x,y,t (internal to lasy) to t,y,x (in openPMD file)
        # This is because many PIC codes expect x to be the fastest index
        data = np.transpose(array).copy()
    else:  # dim == "rt"
        # The representation of modes in openPMD
        # (see https://github.com/openPMD/openPMD-standard/blob/latest/STANDARD.md#required-attributes-for-each-mesh-record)
        # is different than the representation of modes internal to lasy.
        # Thus, there is a non-trivial conversion here
        ncomp = 2 * grid.n_azimuthal_modes - 1
        data = np.zeros((ncomp, grid.npoints[0], grid.npoints[1]), dtype=array.dtype)
        data[0, :, :] = array[0, :, :]
        for mode in range(1, grid.n_azimuthal_modes):
            # cos(m*theta) part of the mode
            data[2 * mode - 1, :, :] = array[mode, :, :] + array[-mode, :, :]
            # sin(m*theta) part of the mode
            data[2 * mode, :, :] = -1.0j * array[mode, :, :] + 1.0j * array[-mode, :, :]
        # Switch from m,r,t (internal to lasy) to m,t,r (in openPMD file)
        # This is because many PIC codes expect r to be the fastest index
        data = np.transpose(data, axes=[0, 2, 1]).copy()

    # Define the dataset
    dataset = io.Dataset(data.dtype, data.shape)
    env = m[io.Mesh_Record_Component.SCALAR]
    env.position = np.zeros(len(dim), dtype=np.float64)
    env.reset_dataset(dataset)
    env.store_chunk(data)

    series.flush()
    series.close()


def extract_array(m, series, component=None):
    """
    Extract and reorder openPMD-formatted array to LASY ordering.

    Namely, ensure that:
     - The longitudinal dimension is t, not z
     - The last axis is t

    Parameters
    ----------
    m : openPMD-api mesh record object
        The array and metadata are read from this.

    series : openPMD Series
        The series containing data m. Only passed for the flush.

    Returns
    -------
    axes_order : List of strings
        Name and ordering of the axes array.
        Input argument for FromArrayProfile, see there for more details.

    axes : Python dictionary containing the axes vectors
        e.g. keys: 'x', 'y', 't' and values: the 1D arrays of each axis.
        Input argument for FromArrayProfile, see there for more details.

    array : 3D array of complex numbers
        Reordered array, with axes in the right order and t last.

    """
    if component is not None:
        array = m[component].load_chunk()
        position = m[component].get_attribute("position")
    else:
        array = m[io.Mesh_Record_Component.SCALAR].load_chunk()
        position = m.get_attribute("position")
    series.flush()
    # node (0.0) or cell (0.5) centered info for each axis
    axis_labels = m.get_attribute("axisLabels")
    grid_offset = m.get_attribute("gridGlobalOffset")
    grid_spacing = m.get_attribute("gridSpacing")
    assert len(axis_labels) in [2, 3]
    if len(axis_labels) == 2:
        idx_offset = 1
        assert axis_labels in [["r", "z"], ["z", "r"], ["r", "t"], ["t", "r"]]
    else:  # len(axis_labels) == 3
        idx_offset = 0
        assert axis_labels in [
            ["x", "y", "z"],
            ["z", "y", "x"],
            ["x", "y", "t"],
            ["t", "y", "x"],
        ]

    # Define parameters to create a profile
    axes = {}
    axes_order = []
    for idx, label in enumerate(axis_labels):
        # Define the axis array
        N = array.shape[idx + idx_offset]
        axis = np.linspace(
            grid_offset[idx] + position[idx] * grid_spacing[idx],
            grid_offset[idx] + (N - 1 + position[idx]) * grid_spacing[idx],
            N,
        )
        # If label is `z`, change it to `t`
        if label == "z":
            axis = (axis - axis[0]) / c
            array = np.flip(array, axis=idx + idx_offset)
            label = "t"

        # Add axis to the dictionary and label to the list
        axes[label] = axis
        axes_order.append(label)

    # Set to LASY order here, time is last axis.
    if axes_order[0] == "t":
        axes_order = axes_order[::-1]
        array = np.swapaxes(array, idx_offset, 2)

    return axes_order, axes, array


def convert_modes(arr_list, geometry, is_env, verbose=False):
    """
    Convert from openPMD mode decomposition to LASY mode decomposition.

    Convert from openPMD mode decomposition of the electric field in cos(m*theta) and sin(m*theta), stored, m in [0, Nmodes] to LASY mode decomposition exp(i*m*theta) m in [-Nmodes+1,Nmodes-1] (array of complex numbers, see https://github.com/LASY-org/lasy/blob/development/README.md):
     - Electromagnetic + cylindrical: we assume Er and Etheta, and construct Ex and Ey.
        https://github.com/openPMD/openPMD-standard/blob/latest/STANDARD.md#required-attributes-for-each-mesh-record. Complex modes, the real and imag part are stored in 2 real arrays.
     - Envelope + cylindrical: we assume the array is E + polarization. Complex modes, stored as arrays of complex numbers. See openPMD link above aas well as https://github.com/openPMD/openPMD-standard/blob/upcoming-2.0.0/EXT_LaserEnvelope.md.
     - Cartesian: do not do anything.

    Parameters
    ----------
    arr_list : list of Numpy arrays
        List of 3D arrays to be converted.

    geometry : string
        Geometry of input data from openPMD standard, "cartesian" or "thetaMode" supported.

    is_env : bool
        Whether the input data represents a laser envelope.
        Otherwise electric field is assumed, specifically x-polarized at the moment.

    verbose : bool (optional)
        If true, print some more intermediate steps.

    Returns
    -------
    arrays_out : list of 3D array
        The arrays converted to LASY mode decomposition.
        This is still the full field, not yet the envelope.
    """
    if geometry == "cartesian":
        return arr_list

    nmodes_in = (arr_list[0].shape[0] + 1) // 2
    if verbose:
        print("nmodes_in:", nmodes_in)
    if is_env:
        assert len(arr_list) == 1
        assert np.iscomplexobj(arr_list[0])
        array_in = arr_list[0]
        array_out = np.zeros_like(arr_list[0], dtype="complex128")
        array_out[0, :, :] = array_in[0, :, :]
        # The data is already Ex, we simply to convert from
        # cos(m*theta) and sin(m*theta) to exp(i*m*theta).
        for imode in range(1, nmodes_in):
            array_out[imode, :, :] = 0.5 * (
                array_in[2 * imode - 1] + 1j * array_in[2 * imode]
            )
            array_out[-imode, :, :] = 0.5 * (
                array_in[2 * imode - 1] - 1j * array_in[2 * imode]
            )
        return [array_out]
    else:
        # arr_list contains 2 elements, Er and Etheta, that need to be
        # combined into Ex. At this point, still operating on full field.
        assert len(arr_list) == 2
        assert np.isrealobj(arr_list[0]) and np.isrealobj(arr_list[1])
        Er_in = arr_list[0]
        Et_in = arr_list[1]

        nmodes_out = nmodes_in - 1
        if verbose:
            print("nmodes_out:", nmodes_in)
        Ex_out = np.zeros(shape=(2 * nmodes_out - 1, *Er_in.shape[1:]))
        Ey_out = np.zeros(shape=(2 * nmodes_out - 1, *Er_in.shape[1:]))
        # The _in arrays have real and imag parts separated, so we add them
        # together by hand
        for imode in range(nmodes_out):
            # input 2*m - 1 and 2*m are real and imag part of mode m, respectively
            # The +1 is conversion from Er & Etheta representation to Ex
            # output exp(i*m*theta) modes: for some reasons, all data goes in
            # m >= 0 modes
            Ex_out[imode, :, :] = 0.5 * (
                Er_in[2 * (imode + 1) - 1, :, :] - Et_in[2 * (imode + 1), :, :]
            )
            Ey_out[imode, :, :] = 0.5 * (
                Er_in[2 * (imode + 1), :, :] + Et_in[2 * (imode + 1) - 1, :, :]
            )
            # Here we assume that mode 0 is only plasma, so we never use it
            if imode >= 2:
                Ex_out[imode, :, :] += 0.5 * (
                    Er_in[2 * (imode - 1) - 1, :, :] + Et_in[2 * (imode - 1), :, :]
                )
                Ey_out[imode, :, :] += 0.5 * (
                    -Er_in[2 * (imode - 1), :, :] + Et_in[2 * (imode - 1) - 1, :, :]
                )
        return [Ex_out, Ey_out]


def isolate_polarization(arrays, dim):
    """
    Extract single array and polarization vector from Ex and Ey.

    Parameters
    ----------
    arrays : list of 2 3D numpy arrays
        Envelopes of Ex and Ey.

    dim : string
        Dimensionality of the array. Options are:

        - 'xyt': The laser pulse is represented on a 3D grid:
                 Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - 'rt' : The laser pulse is represented on a 2D grid:
                 Cylindrical (r) transversely, and temporal (t) longitudinally.

    Returns
    -------
    array : 3D numpy array
        Laser envelope at LASY convention

    pol : tuple of 2 elements
        Polarization vector (px, py), both px and py are complex numbers.
    """
    Ex, Ey = arrays  # Ex and Ey envelopes
    if dim == "rt":
        print(
            "Cylindrical input with full field: polarization is extract from mode 0 only"
        )
        Ex = Ex[0]
        Ey = Ey[0]
    rho2 = np.abs(Ex) ** 2 + np.abs(Ey) ** 2
    # Amplitude of polarization vectors
    rho_x = np.sqrt(np.sum(np.abs(Ex) ** 2) / np.sum(rho2))
    rho_y = np.sqrt(np.sum(np.abs(Ey) ** 2) / np.sum(rho2))
    # Phase in x is assumed 0. This is a convention.
    phi_x = 0
    phi_y = np.average(np.angle(Ey) - np.angle(Ex), weights=rho2)
    px = rho_x * np.exp(1j * phi_x)
    py = rho_y * np.exp(1j * phi_y)
    pol = (px, py)
    print("polarization state detected: (px, py) =", pol)
    array = arrays[0] / px if rho_x >= rho_y else arrays[1] / py

    return array, pol



"""
Test the implementation of spatio-temporal coupling (STC).

This test verifies the correct implementation of initialization and diagnostics for spatio-temporal coupling in Gaussian lasers. It does so by creating a Gaussian pulse at focus and calculating the STC factors using the implemented functions in `laser.utils`.

Additionally, the correctness is validated by comparing the Gaussian profile with a combined Gaussian profile without STC out of focus.
"""

import numpy as np
import scipy.constants as scc

from lasy.laser import Laser
from lasy.profiles.combined_profile import CombinedLongitudinalTransverseProfile
from lasy.profiles.gaussian_profile import GaussianProfile
from lasy.profiles.longitudinal import GaussianLongitudinalProfile
from lasy.profiles.transverse import GaussianTransverseProfile
from lasy.utils.laser_utils import get_beta, get_gdd, get_zeta

wavelength = 0.6e-6  # m
pol = (1, 0)
laser_energy = 1.0  # J
w0 = 5e-6  # m
tau = 5e-14  # s
t_peak = 0.0  # s
beta = 3e-18  # s
zeta = 2.4e-22  # m * s
phi2 = 2.4e-27  # s ^ 2
stc_theta = scc.pi / 2  # rad
z_r = (np.pi * w0**2) / wavelength
z_foc = 3 * z_r

# Create STC profile.
STCprofile = GaussianProfile(
    wavelength=wavelength,
    pol=pol,
    laser_energy=laser_energy,
    w0=w0,
    tau=tau,
    t_peak=t_peak,
    beta=beta,
    zeta=zeta,
    phi2=phi2,
    stc_theta=stc_theta,
)

# Create laser with given profile in `xyt` geometry.
laser_3d = Laser(
    dim="xyt",
    lo=(-10e-6, -10e-6, -30e-14),
    hi=(10e-6, 10e-6, +30e-14),
    npoints=(100, 100, 200),
    profile=STCprofile,
)

# Create laser out of focus with given profile in `rt` geometry by combined and gaussian profile.
long_profile = GaussianLongitudinalProfile(wavelength, tau, t_peak)
trans_profile = GaussianTransverseProfile(w0, wavelength, z_foc)

combined_profile = CombinedLongitudinalTransverseProfile(
    wavelength, pol, long_profile, trans_profile, laser_energy=laser_energy
)

gaussian_profile = GaussianProfile(
    wavelength=wavelength,
    pol=pol,
    laser_energy=laser_energy,
    w0=w0,
    tau=tau,
    t_peak=t_peak,
    z_foc=z_foc,
)

laser_2d_combined = Laser(
    dim="rt",
    lo=(0e-6, -10e-14),
    hi=(50e-6, +10e-14),
    npoints=(60, 200),
    profile=combined_profile,
)

laser_2d_gaussian = Laser(
    dim="rt",
    lo=(0e-6, -10e-14),
    hi=(50e-6, +10e-14),
    npoints=(60, 200),
    profile=gaussian_profile,
)

# calculate the error of gaussian profile
env_combined = laser_2d_combined.grid.get_temporal_field()
env_gaussian = laser_2d_gaussian.grid.get_temporal_field()

err_real = np.average(
    (np.array(env_combined.real) - np.array(env_gaussian.real))
    / np.array(env_combined.real)
)
err_imag = np.average(
    (np.array(env_combined.imag) - np.array(env_gaussian.imag))
    / np.array(env_combined.imag)
)

gdd_3d, gdd0_3d = get_gdd(
    grid=laser_3d.grid, dim=laser_3d.dim, omega0=laser_3d.profile.omega0
)
[zeta_x, zeta_y], [nu_x, nu_y] = get_zeta(
    laser_3d.dim, laser_3d.grid, 2.0 * np.pi / 0.6e-6
)
[beta_x, beta_y] = get_beta(laser_3d.dim, laser_3d.grid, 2.0 * np.pi / 0.6e-6)

assert (err_real + err_imag) < 1e-6

np.testing.assert_approx_equal(gdd0_3d, phi2, significant=2)
np.testing.assert_approx_equal(zeta_y, zeta, significant=2)
np.testing.assert_approx_equal(beta_y, beta, significant=2)



import numpy as np

from .optical_element import OpticalElement


class SpectralFilter(OpticalElement):
    r"""
    Class for an optical element that filters the spectrum of a laser pulse.

    The amplitude multiplier corresponds to:

    .. math::

        T(\omega) = \sqrt{F(\omega)}

    where :math:`F(\omega)` is the intensity transmission function of the filter.

    Parameters
    ----------
    transmission : 1D ndarray of floats
        Intensity/Energy transmission of the filter.
    omega : 1D ndarray of floats (in rad/s)
        Angular frequencies at which the transmission is defined.
    """

    def __init__(self, transmission, omega):
        self.transmission = transmission
        self.omega_in = omega

    def amplitude_multiplier(self, x, y, omega):
        """
        Return the amplitude multiplier.

        Parameters
        ----------
        x, y, omega : ndarrays of floats
            Define points on which to evaluate the multiplier.
            These arrays need to all have the same shape.

        Returns
        -------
        multiplier : ndarray of complex numbers
            Contains the value of the multiplier at the specified points.
        """
        # Sort the transmission function and angular frequencies to
        # eliminate problems with interpolation (e.g. due to fftshifted data)
        order = np.argsort(self.omega_in)
        self.omega_in = self.omega_in[order]
        self.transmission = self.transmission[order]

        # interpolate transmission function to omega axis of the laser
        transmission = np.interp(omega, self.omega_in, self.transmission)

        # return the square root of the energy/intensity transmission
        return np.sqrt(transmission)



import numpy as np

time_axis_indx = -1


class Grid:
    """
    Store the envelope in temporal and spectral space and corresponding metadata.

    Parameters
    ----------
    dim : string
        Dimensionality of the array. Options are:

        - ``'xyt'``: The laser pulse is represented on a 3D grid:
                    Cartesian (x,y) transversely, and temporal (t) longitudinally.
        - ``'rt'`` : The laser pulse is represented on a 2D grid:
                    Cylindrical (r) transversely, and temporal (t) longitudinally.

    lo, hi : list of scalars
        Lower and higher end of the physical domain.
        One element per direction (2 for ``dim='rt'``, 3 for ``dim='xyt'``)

    npoints : tuple of int
        Number of points in each direction.
        One element per direction (2 for ``dim='rt'``, 3 for ``dim='xyt'``)
        For the moment, the lower end is assumed to be (0,0) in rt and (0,0,0) in xyt

    n_azimuthal_modes : int (optional)
        Only used if ``dim`` is ``'rt'``. The number of azimuthal modes
        used in order to represent the laser field.

    is_envelope : bool (optional)
        Whether the field provided uses the (complex) envelope representation, as
        used internally in lasy. If False, field is assumed to represent the
        the full (real) electric field (with fast oscillations).

    is_cw : bool (optional)
        Whether the laser pulse longitudinal profile is a continuous wave laer profile
        or not.

    is_plane_wave : bool (optional)
        Whether the laser pulse transverse profile is a plane wave laer profile or not.
    """

    def __init__(
        self,
        dim,
        lo,
        hi,
        npoints,
        n_azimuthal_modes=None,
        is_envelope=True,
        is_cw=False,
        is_plane_wave=False,
    ):
        # Metadata
        ndims = 2 if dim == "rt" else 3
        assert dim in ["rt", "xyt"]
        assert len(lo) == ndims
        assert len(hi) == ndims

        lo = list(lo)
        hi = list(hi)
        npoints = list(npoints)

        if is_cw:
            if npoints[-1] != 1:
                print(
                    "CW profile: overwrite npoints to only 1 cell in the longitudinal direction."
                )
            lo[-1] = -0.5
            hi[-1] = 0.5
            npoints[-1] = 1
        if is_plane_wave:
            if npoints[0] != 1:
                print(
                    "Plane wave: overwrite npoints to only 1 cell in the transverse directions."
                )
            if dim == "rt":
                lo[0] = 0.0
                hi[0] = np.sqrt(1 / np.pi)
                npoints[0] = 1
            else:
                lo[0] = -0.5
                hi[0] = 0.5
                npoints[0] = 1
                lo[1] = -0.5
                hi[1] = 0.5
                npoints[1] = 1

        self.npoints = npoints
        self.axes = []
        self.dx = []
        for i in range(ndims):
            self.axes.append(np.linspace(lo[i], hi[i], npoints[i]))
            if len(self.axes[i]) > 1:
                self.dx.append(self.axes[i][1] - self.axes[i][0])
            else:
                self.dx.append(hi[i] - lo[i])

        self.lo = lo
        self.hi = hi

        if dim == "rt":
            self.n_azimuthal_modes = n_azimuthal_modes
            self.azimuthal_modes = np.r_[
                np.arange(n_azimuthal_modes), np.arange(-n_azimuthal_modes + 1, 0, 1)
            ]

        # Data
        if dim == "xyt":
            self.shape = self.npoints
        elif dim == "rt":
            # Azimuthal modes are arranged in the following order:
            # 0, 1, 2, ..., n_azimuthal_modes-1, -n_azimuthal_modes+1, ..., -1
            ncomp = 2 * self.n_azimuthal_modes - 1
            self.shape = (ncomp, self.npoints[0], self.npoints[1])

        self.set_is_envelope(is_envelope)
        self.temporal_field = np.zeros(self.shape, dtype=self.dtype)
        self.temporal_field_valid = False
        self.spectral_field = np.zeros(self.shape, dtype="complex128")
        self.spectral_field_valid = False

    def set_is_envelope(self, is_envelope):
        """
        Set is_envelope attribute. Also set dtype accordingly.

        Parameters
        ----------
        is_envelope : boolean
            Whether the grid should represent an envelope (True) or a full electric field (False)
        """
        assert is_envelope in [True, False]
        if is_envelope:
            self.dtype = "complex128"
        else:
            self.dtype = "float64"
        if hasattr(self, "temporal_field"):
            self.temporal_field = self.temporal_field.astype(dtype=self.dtype)
        self.is_envelope = is_envelope

    def set_temporal_field(self, field):
        """
        Set the temporal field.

        Parameters
        ----------
        field : ndarray of complexs
            The temporal field.
        """
        assert field.shape == self.temporal_field.shape
        assert field.dtype == self.dtype
        self.temporal_field[:, :, :] = field
        self.temporal_field_valid = True
        self.spectral_field_valid = False  # Invalidates the spectral field

    def set_spectral_field(self, field):
        """
        Set the spectral field.

        Parameters
        ----------
        field : ndarray of complexs
            The spectral field.
        """
        assert field.shape == self.spectral_field.shape
        assert field.dtype == "complex128"
        self.spectral_field[:, :, :] = field
        self.spectral_field_valid = True
        self.temporal_field_valid = False  # Invalidates the temporal field

    def get_temporal_field(self):
        """
        Return a copy of the temporal field.

        (Modifying the returned object will not modify the original field stored
        in the Grid object ; one must use set_temporal_field to do so.)

        Returns
        -------
        field : ndarray of complexs
            The temporal field.
        """
        # We return a copy, so that the user cannot modify
        # the original field, unless get_temporal_field is called
        if self.temporal_field_valid:
            return self.temporal_field.copy()
        elif self.spectral_field_valid:
            self.spectral2temporal_fft()
            return self.temporal_field.copy()
        else:
            raise ValueError("Both temporal and spectral fields are invalid")

    def get_spectral_field(self):
        """
        Return a copy of the spectral field.

        (Modifying the returned object will not modify the original field stored
        in the Grid object ; one must use set_spectral_field to do so.)

        Returns
        -------
        field : ndarray of complexs
            The spectral field.

        omega : 1d array of real numbers
            The frequency axis consistent with the spectral field.
            This is centered around 0, the central frequency of the envelope
            must be added separately to construct the physical frequency array.
        """
        # We return a copy, so that the user cannot modify
        # the original field, unless set_spectral_field is called
        assert self.is_envelope
        if self.spectral_field_valid:
            return self.spectral_field.copy(), self.spectral_axis.copy()
        elif self.temporal_field_valid:
            self.temporal2spectral_fft()
            return self.spectral_field.copy(), self.spectral_axis.copy()
        else:
            raise ValueError("Both temporal and spectral fields are invalid")

    def temporal2spectral_fft(self):
        """
        Perform the Fourier transform of field from temporal to spectral space.

        (Only along the time axis, not along the transverse spatial coordinates.)
        """
        assert self.temporal_field_valid

        shifted_temporal = np.fft.fftshift(self.temporal_field, axes=time_axis_indx)
        self.spectral_field = np.fft.ifft(shifted_temporal, axis=time_axis_indx)
        self.spectral_axis = 2 * np.pi * np.fft.fftfreq(self.npoints[-1], self.dx[-1])
        self.spectral_field_valid = True

    def spectral2temporal_fft(self):
        """
        Perform the Fourier transform of field from spectral to temporal space.

        (Only along the time axis, not along the transverse spatial coordinates.)
        """
        assert self.spectral_field_valid

        shifted_temporal = np.fft.fft(self.spectral_field, axis=time_axis_indx)
        self.temporal_field = np.fft.fftshift(shifted_temporal, axes=time_axis_indx)

        self.temporal_field_valid = True
